import { ResolutionError, CircularDependencyError } from '../exceptions/index.js';
import { owner, scope } from '../common/index.js';
export class InternalContainer {
    constructor(_registrationStorageService, _disposableStorageService, _registrationBaseService, _containerApiService, _invokerService, _resolutionCacheService, _resolutionDetails) {
        this._registrationStorageService = _registrationStorageService;
        this._disposableStorageService = _disposableStorageService;
        this._registrationBaseService = _registrationBaseService;
        this._containerApiService = _containerApiService;
        this._invokerService = _invokerService;
        this._resolutionCacheService = _resolutionCacheService;
        this._resolutionDetails = _resolutionDetails;
        this._children = [];
        this._dependencyScope = scope.none;
        this._dependencyOwner = owner.externals;
        this._collection = this._registrationStorageService.create();
        this._disposableStorage = this._disposableStorageService.create();
        this._invoker = this._invokerService.create(this, _resolutionDetails);
        this._pendingResolutions = {};
        this.registerImpl = this.registerImpl.bind(this);
        this._resolutionCache = _resolutionCacheService.create();
        this._cache = this.createCache(this._resolutionCache);
    }
    get cache() {
        return this._cache;
    }
    add(registrations) {
        registrations.forEach(this.registerImpl);
    }
    createChild() {
        const child = new InternalContainer(this._registrationStorageService, this._disposableStorageService, this._registrationBaseService, this._containerApiService, this._invokerService, this._resolutionCacheService, this._resolutionDetails);
        child._parent = this;
        this._children.push(child);
        return child;
    }
    dispose() {
        this._disposableStorage.disposeItems();
        while (this._children.length > 0) {
            const item = this._children.pop();
            item.dispose();
        }
        this._collection.clear();
        this._resolutionCache.clear();
    }
    disposeAsync() {
        throw new Error('Not implemented');
    }
    resolve(service, ...args) {
        const registration = this.createRegistration(service);
        registration.args = args;
        return this.resolveBase(registration, true);
    }
    resolveAsync(_service, ..._args) {
        throw new Error('Not implemented');
    }
    tryResolve(service, ...args) {
        const registration = this.createRegistration(service);
        registration.args = args;
        return this.resolveBase(registration, false);
    }
    tryResolveAsync(_service, ..._args) {
        throw new Error('Not implemented');
    }
    resolveNamed(service, name, ...args) {
        const registration = this.createRegistration(service);
        registration.name = name;
        registration.args = args;
        return this.resolveBase(registration, true);
    }
    resolveNamedAsync(_service, _name, ..._args) {
        throw new Error('Not implemented');
    }
    tryResolveNamed(service, name, ...args) {
        const registration = this.createRegistration(service);
        registration.name = name;
        registration.args = args;
        return this.resolveBase(registration, false);
    }
    tryResolveNamedAsync(_service, _name, ..._args) {
        throw new Error('Not implemented');
    }
    resolveWithDependencies(service, dependencies) {
        // Import API container is passed as undefined because we only
        // care about carrying the dependencies
        const api = this._containerApiService.create(undefined);
        api.service(service)
            .dependencies(dependencies);
        return this.resolveWithDepBase(api);
    }
    resolveWithDependenciesAsync(_service, _dependencies) {
        throw new Error('Not implemented');
    }
    resolveWith(service) {
        const importApi = {
            execute: (api) => {
                let result;
                if (api.isDependenciesResolvable) {
                    result = this.resolveWithDepBase(api);
                }
                else {
                    const registration = this.createRegistration(api.serviceValue);
                    registration.name = api.nameValue;
                    registration.args = api.argsValue;
                    result = this.resolveBase(registration, api.throwResolveError);
                }
                if (result && api.cacheValue.use) {
                    this.addToCache(result, api);
                }
                return result;
            }
        };
        const api = this._containerApiService.create(importApi);
        return api.service(service);
    }
    registerImpl(registration) {
        registration.checkRegistrationType();
        registration.container = this;
        this._collection.addEntry(registration);
    }
    resolveBase(registration, throwIfNotFound) {
        const entry = this.resolveImpl(registration, throwIfNotFound);
        if (!entry && !throwIfNotFound) {
            return undefined;
        }
        if (this._pendingResolutions[entry.id]) {
            throw new CircularDependencyError(`${entry.service}`);
        }
        try {
            this._pendingResolutions[entry.id] = true;
            // ------------ with args always returns new instance ...
            if (registration.args && registration.args.length) {
                return this.createTrackable(entry, throwIfNotFound, registration.args);
            }
            return this.resolveScope(entry, throwIfNotFound);
        }
        finally {
            delete this._pendingResolutions[entry.id];
        }
    }
    resolveImpl(registration, throwIfNotFound) {
        const serviceEntry = this._collection.getEntry(registration);
        if (!serviceEntry && this._parent) {
            return this._parent.resolveImpl(registration, throwIfNotFound);
        }
        if (!serviceEntry && throwIfNotFound) {
            throw new ResolutionError({
                message: 'Could not resolve service',
                data: registration.service
            });
        }
        return serviceEntry;
    }
    resolveScope(registration, throwIfNotFound) {
        switch (registration.scope) {
            case scope.none:
                return this.createTrackable(registration, throwIfNotFound);
            case scope.container:
                return this.resolveContainerScope(registration, throwIfNotFound);
            case scope.hierarchy:
                return this.resolveHierarchyScope(registration, throwIfNotFound);
            default:
                throw new ResolutionError({ message: 'Unknown scoping' });
        }
    }
    resolveContainerScope(registration, throwIfNotFound) {
        let entry;
        if (registration.container !== this) {
            entry = registration.cloneFor(this);
            this._collection.addEntry(entry);
        }
        else {
            entry = registration;
        }
        if (!entry.instance) {
            entry.instance = this.createTrackable(entry, throwIfNotFound);
        }
        return entry.instance;
    }
    resolveHierarchyScope(registration, throwIfNotFound) {
        if (registration.container &&
            registration.container !== this) {
            const container = registration.container;
            return container.resolveBase(registration, throwIfNotFound);
        }
        if (!registration.instance) {
            registration.instance = this.createTrackable(registration, throwIfNotFound);
        }
        return registration.instance;
    }
    createTrackable(registration, throwIfNotFound, args) {
        try {
            let instance = this._invoker.invoke(registration, throwIfNotFound, args);
            if (registration.initializer) {
                instance = registration.initializer(this, instance);
            }
            if (registration.owner === owner.container &&
                !registration.isLazy &&
                registration.disposer) {
                this._disposableStorage.add(instance, registration.disposer);
            }
            return instance;
        }
        catch (error) {
            if (error instanceof CircularDependencyError) {
                throw error;
            }
            let message = 'Could not instantiate service';
            if (error instanceof ResolutionError) {
                message += `. ${error.message}`;
            }
            throw new ResolutionError({ message, data: registration.service, error });
        }
    }
    createRegistration(service) {
        return this._registrationBaseService.create(service);
    }
    createDependenciesRegistration(api) {
        const items = api.dependenciesValue.map(dependency => {
            if (!dependency.service) {
                throw new ResolutionError({ message: 'Service is not defined', data: dependency });
            }
            const registration = this.createRegistration(dependency.service);
            registration.copyDependency(dependency);
            registration.checkRegistrationType();
            const throwOnError = dependency.required !== false && api.throwResolveError;
            return {
                implementation: this.resolveImpl(registration, throwOnError),
                dependency
            };
        })
            .filter(item => item.implementation ||
            item.dependency.required === false ? true : false);
        if (items.length !== api.dependenciesValue.length)
            return [];
        return items.map(item => {
            const baseRegistration = item.dependency.required === false ?
                this.createRegistration(item.dependency.service)
                : item.implementation.cloneFor(this);
            baseRegistration.copyDependency(item.dependency);
            baseRegistration.disposer = undefined;
            baseRegistration.scope = this._dependencyScope;
            baseRegistration.owner = this._dependencyOwner;
            return baseRegistration;
        });
    }
    resolveWithDepBase(api) {
        const child = this.createChild();
        const registration = this.createRegistration(api.serviceValue);
        registration.args = api.argsValue;
        registration.name = api.nameValue;
        const implementation = this.resolveImpl(registration, api.throwResolveError);
        if (!implementation && !api.throwResolveError) {
            return undefined;
        }
        const baseRegistration = implementation.cloneFor(child);
        baseRegistration.args = api.argsValue;
        baseRegistration.name = api.nameValue;
        baseRegistration.disposer = undefined;
        baseRegistration.dependenciesValue = api.dependenciesValue;
        baseRegistration.scope = this._dependencyScope;
        baseRegistration.owner = this._dependencyOwner;
        const registrations = this.createDependenciesRegistration(api);
        if (registrations.length <= 0) {
            return undefined;
        }
        registrations.push(baseRegistration);
        child.add(registrations);
        return child.resolveBase(baseRegistration, api.throwResolveError);
    }
    createCache(resolutionCache) {
        return {
            get instance() {
                return resolutionCache.instance;
            },
            resolve(name) {
                return resolutionCache.resolve(name);
            }
        };
    }
    addToCache(value, api) {
        let name;
        if (api.cacheValue.name) {
            name = api.cacheValue.name;
        }
        else if (api.nameValue) {
            name = api.nameValue;
        }
        else if (api.serviceValue.name) {
            name = api.serviceValue.name;
        }
        else if (typeof api.serviceValue === 'string') {
            name = api.serviceValue;
        }
        else {
            throw new ResolutionError({ message: 'Missing cache name' });
        }
        this._resolutionCache.add(name, value);
    }
}
