import { registrationType } from '../registration/index.js';
import { decoratorResolutionParameter } from '../decorators/index.js';
import { ResolutionError } from '../exceptions/index.js';
import { construct, getMetadata } from '../utils/index.js';
export class Invoker {
    constructor(_container, _resolutionDetails) {
        this._container = _container;
        this._resolutionDetails = _resolutionDetails;
    }
    invoke(registration, throwIfNotFound, args) {
        if (registration.isLazy) {
            return () => this.invokeInternal(registration, throwIfNotFound, args);
        }
        return this.invokeInternal(registration, throwIfNotFound, args);
    }
    invokeInternal(registration, throwIfNotFound, args) {
        switch (registration.registrationType) {
            case registrationType.factoryType:
                return this.instantiate(registration.factoryType, registration, throwIfNotFound, args);
            case registrationType.factoryValue:
                return registration.factoryValue;
            case registrationType.factory:
            default:
                return this.createByFactory(registration, args);
        }
    }
    createByFactory(registration, args = []) {
        return registration.factory(registration.container, ...args);
    }
    instantiate(type, registration, throwIfNotFound, args) {
        if (args && args.length &&
            registration.params.length) {
            throw new ResolutionError({
                message: [
                    'Could not instantiate type.',
                    'Arguments and dependencies are not allowed for simultaneous resolution.',
                    'Pick dependencies or arguments'
                ].join(' '),
                data: type
            });
        }
        if (args && args.length) {
            return construct(type, args);
        }
        if (registration.params.length) {
            return this.instantiateByParams(type, registration, throwIfNotFound);
        }
        const dependencies = this.getDependencies(type);
        return this.instantiateByDependencies(type, dependencies);
    }
    instantiateByParams(type, registration, throwIfNotFound) {
        const params = registration.params
            .map(item => {
            const dependency = registration.dependenciesValue.filter(d => d.service === item)[0];
            const dependencyName = dependency ? dependency.named : null;
            if (throwIfNotFound) {
                return !!dependencyName ?
                    this._container.resolveNamed(item, dependencyName) :
                    this._container.resolve(item);
            }
            return !!dependencyName ?
                this._container.tryResolveNamed(item, dependencyName) :
                this._container.tryResolve(item);
        })
            .filter(item => item);
        return construct(type, params);
    }
    instantiateByDependencies(type, dependencies) {
        const params = dependencies
            .map((dependency, index) => {
            const depParams = this._resolutionDetails ? this._resolutionDetails.tryGet(type) : null;
            const depParamsValue = depParams ? depParams[index] : null;
            if (!depParamsValue) {
                return this._container.resolve(dependency);
            }
            if (depParamsValue.type === decoratorResolutionParameter.value) {
                return depParamsValue.value;
            }
            if (depParamsValue.type === decoratorResolutionParameter.functionValue) {
                return depParamsValue.value();
            }
            const resolutionItem = depParamsValue.service || dependency;
            const resolution = this._container.resolveWith(resolutionItem);
            if (depParamsValue.args && depParamsValue.args.length) {
                resolution.args(...depParamsValue.args);
            }
            if (depParamsValue.name) {
                resolution.name(depParamsValue.name);
            }
            if (depParamsValue.attempt === true) {
                resolution.attempt();
            }
            if (depParamsValue.cache && depParamsValue.cache.use) {
                resolution.cache(depParamsValue.cache.name);
            }
            return resolution.exec();
        });
        return construct(type, params);
    }
    getDependencies(type) {
        return getMetadata(type);
    }
}
