import { DecoratorError } from '../exceptions/index.js';
import { decoratorResolutionParameter } from './types/index.js';
import { defaults } from '../common/index.js';
import { getMetadata, isPrototype, isFunction } from '../utils/index.js';
export class Decorator {
    constructor(builderService, internalContainerService, _decoratorRegistrationApiService, _internalStorageService) {
        this._decoratorRegistrationApiService = _decoratorRegistrationApiService;
        this._internalStorageService = _internalStorageService;
        this._internalStorage = this._internalStorageService.create();
        internalContainerService.resolutionDetails = this._internalStorage;
        this._builder = builderService.create(internalContainerService);
    }
    build() {
        return this._builder.build();
    }
    provide(service) {
        const register = createRegister(this._builder);
        const api = this._decoratorRegistrationApiService.createRegistration(register);
        return api.provide(service);
    }
    provideSelf() {
        const register = createRegister(this._builder, true);
        const api = this._decoratorRegistrationApiService.createRegistration(register);
        return api.provideUndefined();
    }
    by(service) {
        const resolve = (api) => (target, _key, index) => {
            if (!api.service) {
                const dependencies = getMetadata(target);
                api.service = dependencies[index];
            }
            const bucket = this._internalStorage.register(target, () => ({}));
            bucket[index] = {
                service: api.service,
                args: api.args,
                attempt: api.attempt,
                name: api.name,
                cache: api.cache,
                type: decoratorResolutionParameter.service
            };
        };
        const api = this._decoratorRegistrationApiService.createResolution(resolve);
        return api.by(service);
    }
    resolveValue(value) {
        return (target, _key, index) => {
            const bucket = this._internalStorage.register(target, () => ({}));
            const type = isFunction(value) ?
                decoratorResolutionParameter.functionValue :
                decoratorResolutionParameter.value;
            bucket[index] = {
                value,
                type
            };
        };
    }
    register(service) {
        return this._builder.register(service);
    }
    import(builder) {
        this._builder.copy(builder);
    }
}
function createRegister(builder, isSelf = false) {
    return (api) => (target) => {
        if (!isPrototype(target)) {
            throw new DecoratorError({ message: 'Decorator target not supported, not a prototype', data: { target } });
        }
        const registration = isSelf ?
            builder.register(target).asSelf() :
            builder.register(api.service).asType(target);
        const initializer = api.initializedBy;
        if (initializer) {
            registration.initializeBy(initializer);
        }
        const isLazy = api.isLazy;
        if (isLazy) {
            registration.lazy();
        }
        const disposer = api.disposedBy;
        if (disposer) {
            registration.dispose(disposer);
        }
        const name = api.name;
        if (name) {
            registration.named(name);
        }
        const scope = api.scope || defaults.scope;
        registration.within(scope);
        return target;
    };
}
