import { ImmutableArray, isArray } from '../utils/index.js';
import { callInfo } from '../common/index.js';
import { propertyType } from './types/index.js';
export class Decorator {
    wrap(strategyInfo) {
        strategyInfo = this.copyStrategy(strategyInfo);
        const strategyStore = strategyInfo.substitute ?
            this.defineWrapStrategies() :
            this.defineNonWrapStrategies();
        const strategy = strategyStore[strategyInfo.type];
        strategy(strategyInfo);
    }
    defineNonWrapStrategies() {
        const result = {};
        result[propertyType.method] = (strategyInfo) => {
            const value = strategyInfo.source[strategyInfo.name];
            strategyInfo.destination[strategyInfo.name] = function () {
                const args = Array.prototype.slice.call(arguments, 0);
                return value.apply(this, args);
            };
        };
        result[propertyType.getter] = (strategyInfo) => {
            const { configurable, enumerable } = strategyInfo.descriptor ?
                strategyInfo.descriptor : { configurable: true, enumerable: true };
            Object.defineProperty(strategyInfo.destination, strategyInfo.name, {
                get: this.defineGetter(strategyInfo),
                configurable,
                enumerable
            });
        };
        result[propertyType.setter] = (strategyInfo) => {
            const { configurable, enumerable } = strategyInfo.descriptor ?
                strategyInfo.descriptor : { configurable: true, enumerable: true };
            Object.defineProperty(strategyInfo.destination, strategyInfo.name, {
                set: this.defineSetter(strategyInfo),
                configurable,
                enumerable
            });
        };
        result[propertyType.fullProperty] = (strategyInfo) => {
            const { configurable, enumerable } = strategyInfo.descriptor ?
                strategyInfo.descriptor : { configurable: true, enumerable: true };
            Object.defineProperty(strategyInfo.destination, strategyInfo.name, {
                get: this.defineGetter(strategyInfo),
                set: this.defineSetter(strategyInfo),
                configurable,
                enumerable
            });
        };
        result[propertyType.field] = result[propertyType.fullProperty];
        return result;
    }
    defineWrapStrategies() {
        const result = {};
        const createCallChainFromList = this.createCallChainFromList.bind(this);
        const defineWrapGetter = this.defineWrapGetter.bind(this);
        const defineWrapSetter = this.defineWrapSetter.bind(this);
        const defineGetter = this.defineGetter.bind(this);
        const defineSetter = this.defineSetter.bind(this);
        result[propertyType.method] = (strategyInfo) => {
            const value = strategyInfo.source[strategyInfo.name];
            strategyInfo.destination[strategyInfo.name] = function () {
                const args = Array.prototype.slice.call(arguments, 0);
                const delegate = (args) => {
                    if (!args || !isArray(args)) {
                        args = [args];
                    }
                    return value.apply(this, args);
                };
                return createCallChainFromList({
                    delegate,
                    strategyInfo,
                    args: ImmutableArray.createImmutable(args),
                    wrapperContext: this,
                });
            };
        };
        result[propertyType.getter] = (strategyInfo) => {
            Object.defineProperty(strategyInfo.destination, strategyInfo.name, {
                get: defineWrapGetter(strategyInfo),
                configurable: true,
                enumerable: strategyInfo.descriptor ? strategyInfo.descriptor.enumerable : true
            });
        };
        result[propertyType.setter] = (strategyInfo) => {
            Object.defineProperty(strategyInfo.destination, strategyInfo.name, {
                set: defineWrapSetter(strategyInfo),
                configurable: true,
                enumerable: strategyInfo.descriptor ? strategyInfo.descriptor.enumerable : true
            });
        };
        result[propertyType.fullProperty] = (strategyInfo) => {
            const getter = strategyInfo.substitute.type === callInfo.any ||
                strategyInfo.substitute.type === callInfo.getterSetter ||
                strategyInfo.substitute.type === callInfo.getter ||
                strategyInfo.substitute.type === callInfo.field ?
                defineWrapGetter(strategyInfo) : defineGetter(strategyInfo);
            const setter = strategyInfo.substitute.type === callInfo.any ||
                strategyInfo.substitute.type === callInfo.getterSetter ||
                strategyInfo.substitute.type === callInfo.setter ||
                strategyInfo.substitute.type === callInfo.field ?
                defineWrapSetter(strategyInfo) : defineSetter(strategyInfo);
            Object.defineProperty(strategyInfo.destination, strategyInfo.name, {
                get: getter,
                set: setter,
                configurable: true,
                enumerable: strategyInfo.descriptor ? strategyInfo.descriptor.enumerable : true
            });
        };
        result[propertyType.field] = result[propertyType.fullProperty];
        return result;
    }
    defineWrapSetter(strategyInfo) {
        const createCallChainFromList = this.createCallChainFromList.bind(this);
        const defineSetter = this.defineSetter.bind(this);
        return function (value) {
            const delegate = defineSetter(strategyInfo).bind(this);
            return createCallChainFromList({
                delegate,
                strategyInfo,
                args: ImmutableArray.createImmutable([value]),
                wrapperContext: this,
                callType: callInfo.setter,
            });
        };
    }
    defineWrapGetter(strategyInfo) {
        const createCallChainFromList = this.createCallChainFromList.bind(this);
        const defineGetter = this.defineGetter.bind(this);
        return function () {
            const delegate = defineGetter(strategyInfo).bind(this);
            return createCallChainFromList({
                delegate,
                strategyInfo,
                args: ImmutableArray.createImmutable([]),
                wrapperContext: this,
                callType: callInfo.getter,
            });
        };
    }
    defineGetter(strategyInfo) {
        return function () {
            return strategyInfo.contextName ? this[strategyInfo.contextName][strategyInfo.name]
                : strategyInfo.source[strategyInfo.name];
        };
    }
    defineSetter(strategyInfo) {
        return function (argValue) {
            if (strategyInfo.contextName) {
                this[strategyInfo.contextName][strategyInfo.name] = argValue;
            }
            else {
                strategyInfo.source[strategyInfo.name] = argValue;
            }
        };
    }
    createCallChainFromList(info) {
        const mainCallInfo = this.createCallInfo(info);
        this.createCallAction(mainCallInfo, info, info.strategyInfo.substitute.next);
        return info.strategyInfo.substitute.wrapper.call(info.wrapperContext, mainCallInfo);
    }
    createCallAction(callInfo, info, substitute) {
        if (!substitute) {
            return;
        }
        const createCallAction = this.createCallAction.bind(this);
        const childCallInfo = this.createCallInfo(info);
        const wrapper = substitute.wrapper.bind(info.wrapperContext);
        callInfo.next = result => {
            childCallInfo.result = result;
            createCallAction(childCallInfo, info, substitute.next);
            return wrapper(childCallInfo);
        };
    }
    createCallInfo(info) {
        const getter = info.delegate;
        const setter = info.delegate;
        return {
            source: info.strategyInfo.source,
            name: info.strategyInfo.name,
            args: info.args.value,
            type: info.callType || info.strategyInfo.substitute.type,
            invoke: info.delegate,
            get: info.callType === callInfo.getter ? getter : undefined,
            set: info.callType === callInfo.setter ? setter : undefined
        };
    }
    copyStrategy(strategyInfo) {
        return {
            type: strategyInfo.type,
            descriptor: strategyInfo.descriptor,
            substitute: strategyInfo.substitute,
            name: strategyInfo.name,
            source: strategyInfo.source,
            destination: strategyInfo.destination,
            contextName: strategyInfo.contextName
        };
    }
}
