import { callInfo } from '../common/index.js';
import { SubstituteStorage } from './substitute-storage.js';
import { checkNullArgument, isPrototype, isObject, isArray } from '../utils/index.js';
import { ArgumentError } from '../exceptions/index.js';
export class Interceptor {
    constructor(_proxy) {
        this._proxy = _proxy;
    }
    interceptPrototype(subject, substitutes) {
        this.assertPrototype(subject);
        const storage = this.convertParams(subject, substitutes);
        return this._proxy.byPrototype(subject, storage);
    }
    interceptInstance(subject, substitutes) {
        this.assertObject(subject);
        const storage = this.convertParams(subject, substitutes);
        return this._proxy.byInstance(subject, storage);
    }
    intercept(subject, substitutes) {
        const storage = this.convertParams(subject, substitutes);
        return this.convertToIntercept(subject, storage);
    }
    withSubstitute(substitute) {
        const storage = new SubstituteStorage();
        const interceptInstance = (subject) => {
            this.assertObject(subject);
            return this._proxy.byInstance(subject, storage);
        };
        const interceptPrototype = (subject) => {
            this.assertPrototype(subject);
            return this._proxy.byPrototype(subject, storage);
        };
        const intercept = (subject) => {
            return this.convertToIntercept(subject, storage);
        };
        return this.with(interceptInstance, interceptPrototype, intercept, storage, substitute);
    }
    with(interceptInstance, interceptPrototype, intercept, storage, substitute) {
        storage.add(this.createSubstitute(substitute));
        return {
            withSubstitute: this.with.bind(this, interceptInstance, interceptPrototype, intercept, storage),
            interceptInstance,
            interceptPrototype,
            intercept
        };
    }
    convertParams(subject, substitutes) {
        checkNullArgument(subject, 'subject');
        if (!substitutes) {
            return undefined;
        }
        const data = isArray(substitutes) ?
            substitutes :
            [substitutes];
        return this.transformSubstitutes(data);
    }
    convertToIntercept(subject, storage) {
        let result;
        if (isPrototype(subject)) {
            result = this._proxy.byPrototype(subject, storage);
        }
        else if (isObject(subject)) {
            result = this._proxy.byInstance(subject, storage);
        }
        else {
            throw new ArgumentError('subject', {
                message: 'Subject should be a prototype function or an object'
            });
        }
        return result;
    }
    transformSubstitutes(substitutes) {
        const storage = new SubstituteStorage();
        return substitutes.reduce((storage, current) => {
            const substitute = this.createSubstitute(current);
            storage.add(substitute);
            return storage;
        }, storage);
    }
    createSubstitute(value) {
        if (!value.wrapper) {
            throw new ArgumentError('wrapper', {
                message: 'Missing interceptor wrapper', data: value
            });
        }
        return {
            method: value.method,
            type: value.type || callInfo.any,
            wrapper: value.wrapper,
            next: undefined
        };
    }
    assertPrototype(subject) {
        if (!isPrototype(subject)) {
            throw new ArgumentError('subject', {
                message: 'Subject should be a prototype function'
            });
        }
    }
    assertObject(subject) {
        if (isPrototype(subject)) {
            throw new ArgumentError('subject', { message: 'Subject should be an object' });
        }
    }
}
