import { callInfo } from '../common/index.js';
import { ProxyError } from '../exceptions/index.js';
export class SubstituteStorage {
    constructor() {
        this._known = Object.create(null);
        this._unknown = { head: null, tail: null };
    }
    add(value) {
        if (value.method) {
            this.addKnownSubstitute(value);
        }
        else {
            this.addUnknownSubstitute(value);
        }
    }
    getSubstitutes(name, types) {
        const unknown = this._unknown.head ?
            this.copyList(this._unknown, (value) => !value.type || types.indexOf(value.type) >= 0)
            :
                null;
        const bucket = this._known[name];
        if (!bucket) {
            return unknown ? unknown.head : null;
        }
        const result = this.copyList(bucket, (value) => {
            if (!value.type || types.indexOf(value.type) >= 0) {
                return true;
            }
            throw this.createIncompatibleTypeError(value.type, types);
        });
        result.tail.next = (unknown ? unknown.head : result.tail.next) || undefined;
        return result.head;
    }
    addKnownSubstitute(value) {
        const method = value.method;
        const bucket = this._known[method] || { head: null, tail: null };
        this.addToList(value, bucket);
        this._known[method] = bucket;
    }
    addUnknownSubstitute(value) {
        this.addToList(value, this._unknown);
    }
    createIncompatibleTypeError(type, types) {
        const typeMessage = types.map(item => this.callTypeToString(item)).join(', ');
        const message = [
            'Could not match proxy type and property type.',
            `Expected one of: ${typeMessage}.`,
            `Actual: ${this.callTypeToString(type)}`
        ].join(' ');
        throw new ProxyError({ message });
    }
    addToList(value, list) {
        if (!list.head) {
            list.head = value;
            list.tail = value;
        }
        else {
            list.tail.next = value;
            list.tail = value;
        }
    }
    copyList(list, predicate) {
        const result = {
            head: null,
            tail: null
        };
        let current = list.head;
        while (current) {
            if (predicate(current)) {
                const substitute = {
                    method: current.method,
                    type: current.type,
                    wrapper: current.wrapper
                };
                this.addToList(substitute, result);
            }
            current = current.next || null;
        }
        return result;
    }
    callTypeToString(type) {
        switch (type) {
            case callInfo.field: return 'Field';
            case callInfo.getter: return 'Getter';
            case callInfo.setter: return 'Setter';
            case callInfo.getterSetter: return 'GetterSetter';
            case callInfo.method: return 'Method';
            default:
                return 'Any';
        }
    }
}
