/**
 * Represents resolution life-cycle type values
 * @public
 */
declare type ScopeType = 1 | 2 | 3;
/**
 * Represents resolution life-cycle type definition
 * @public
 */
declare const scope: Readonly<{
    /**
     * Transient life-cycle, one instance per resolution
     */
    none: ScopeType;
    /**
     * Container life-cycle, one instance per resoluton per container instance
     */
    container: ScopeType;
    /**
     * Hierarchy life-cycle, same instance per all resolutions
     */
    hierarchy: ScopeType;
}>;
//# sourceMappingURL=scope.d.ts.map

/**
 * Represents member substitution type values
 * @public
 */
declare type CallInfoType = 1 | 2 | 3 | 4 | 5 | 6;
/**
 * Represents original substituted subject member type
 * @public
 */
declare const callInfo: Readonly<{
    /**
     * Indicates method member type substitution
     */
    method: CallInfoType;
    /**
     * Indicates getter member type substitution
     */
    getter: CallInfoType;
    /**
     * Indicates setter member type substitution
     */
    setter: CallInfoType;
    /**
     * Indicates substitution for both getter and setter member types
     */
    getterSetter: CallInfoType;
    /**
     * Indicates substitution of any member type
     */
    any: CallInfoType;
    /**
     * Indicates field substitution
     */
    field: CallInfoType;
}>;
//# sourceMappingURL=call-info-type.d.ts.map

declare type Omit<T, K extends keyof T> = Pick<T, Exclude<keyof T, K>>;//# sourceMappingURL=common.d.ts.map

/**
 * Represents factory method registration interface.
 * Receives an instance of a container {@link IContainer} and
 * an array of optional parameters provided during resolution
 * @public
 */
declare type Factory<T> = (c: IContainer, ...args: any[]) => T;
/**
 * Represents dynamic dependency interface
 * @public
 */
interface IDynamicDependency {
    /**
     * Specifies service value, registration to be substituted for
     */
    service: {};
    /**
     * Specifies factory interface {@link Factory} to be used as substitution for factory
     * registrations
     */
    factory?: Factory<{}>;
    /**
     * Specifies factory type to be used as substitution for factory type
     * registrations
     */
    factoryType?: {};
    /**
     * Specifies factory value to be used as substitution for factory value
     * registrations
     */
    factoryValue?: {};
    /**
     * Specifies a name to be used as substitution for named registrations
     */
    named?: string;
    /**
     * Specifies an instance of initializer {@link Initializer} to be used
     * during substitution initialization
     */
    initializer?: Initializer<{}>;
    /**
     * Specifies the necessity of the registration presence prior substitution.
     * If required is true and no registration found, throws {@link ResolutionError}
     * during resolution
     */
    required?: boolean;
}
/**
 * Specifies an instance of the initialization action to be used during resolution instantiation.
 * Receives an instance of a container {@link IContainer} and an instance of the resolved service
 * @public
 */
declare type Initializer<T> = (c: IContainer, item: T) => T;
/**
 * Specifies an instance of the dispose action to be used during resolution disposal
 * Receives an instance of the resolution
 * @public
 */
declare type Disposer<T> = (item: T) => void;
/**
 * Represents an entry step within registration fluent cascading API sequence
 * @public
 */
interface IRegisterWithAs<T> {
    /**
     * Specifies initialization action during service instantiation
     * @param action - an instance of {@link Initializer} interface
     * @returns - an instance of {@link RegisterWithInitializeBy} interface
     */
    initializeBy<K extends T>(action: Initializer<K>): RegisterWithInitializeBy<K>;
    /**
     * Specifies lazy registration. Resolution of services is differed until results
     * are needed by other services/computations
     * @returns - an instance of {@link RegisterWithLazy} interface
     */
    lazy(): RegisterWithLazy<T>;
    /**
     * Specifies named registration. Registration can be resolved only using the name
     * provided
     * @param name - a value that should be used during service resolution
     * @returns - an instance of {@link RegisterWithName} interface
     */
    named(name: string): RegisterWithName<T>;
    /**
     * Specifies disposer action during resolution disposal.
     * @param action - disposer action to be called during resolution
     * disposal process ({@link Disposer})
     * @returns - an instance of {@link RegisterWithLazy} interface
     */
    dispose<K extends T>(action: Disposer<K>): RegisterWithLazy<K>;
    /**
     * Specifies resolution instance scope. Scope determines how resolved service
     * instances behave in between resolutions
     * @param scope - the value of the scope {@link ScopeType}
     * @returns - void
     */
    within(scope: ScopeType): void;
    /**
     * Specifies transient scope resolution. Every resolution of service returns new instance
     * Is similar to scope.none {@link ScopeType} (default behavior)
     * @returns - void
     */
    transient(): void;
    /**
     * Specifies singleton scope resolution. Every resolution of service returns same instance
     * Is similar to scope.hierarchy {@link scope}
     * @returns - void
     */
    singleton(): void;
    /**
     * Specifies instance per container scope resolution.
     * Every resolution of a service returns the same instance per instance of a container
     * Is similar to scope.container {@link scope}
     * @returns - void
     */
    instancePerContainer(): void;
}
/**
 * Specifies named registration interface with no additional steps
 * within fluent cascading API
 * Receives a value to be used named registrations / resolutions
 * @public
 */
interface IName {
    named(name: string): void;
}
/**
 * Represents a step within fluent cascading API registration sequence where `initializeBy` method
 * was omitted
 * @public
 */
declare type RegisterWithInitializeBy<T> = Omit<IRegisterWithAs<T>, 'initializeBy'>;
/**
 * Represents a step within fluent cascading API registration sequence where `lazy` and
 * `dispose` methods were omitted
 * @public
 */
declare type RegisterWithLazy<T> = Omit<RegisterWithInitializeBy<T>, 'lazy' | 'dispose'>;
/**
 * Represents a step within fluent cascading API registration sequence where `named` method
 * was omitted
 * @public
 */
declare type RegisterWithName<T> = Omit<RegisterWithLazy<T>, 'named'>;
/**
 * Represents an entry into service registration fluent cascading API interface
 * @public
 */
interface IRegistration<T> {
    /**
     * Sets registration to be marked as factory registration
     *
     * @remarks
     * Registration is provided as a factory method
     *
     * @param factory - an instance of {@link Factory} interface
     * @returns - an instance of {@link IRegisterWithAs} interface
     */
    as(factory: Factory<T>): IRegisterWithAs<T>;
    /**
     * Sets registration to be marked as type registration
     *
     * @remarks
     * Registration is provided as a type. An instance of the type is instantiated
     * during resolution. It is intended for construct-able types only. Uses new
     * operator for type construction
     *
     * @param type - type representative to be used for the instance construction
     * @param params - optional array of parameters to be used for the type
     * instance construction
     * @returns - an instance of {@link IRegisterWithAs} interface
     */
    asType(type: T, ...params: any[]): IRegisterWithAs<T>;
    /**
     * Sets registration to be marked as self registration
     *
     * @remarks
     * Registration is provided as a type itself (participates both in registration and
     * resolution). It is intended for construct-able types only. Uses new
     * operator for type construction
     *
     * @param params - optional array of parameters to be used for the type
     * @returns - an instance of {@link IRegisterWithAs} interface
     */
    asSelf(...params: any[]): IRegisterWithAs<T>;
    /**
     * Sets registration to be marked as value registration
     *
     * @remarks
     * Registration is provided as a value that is returned during resolution
     *
     * @param value - value to be returned during resolution
     * @returns - an instance of {@link IName} interface
     */
    asValue(value: {}): IName;
}
//# sourceMappingURL=common.d.ts.map

/**
 * Represents cached resolutions interface.
 *
 * @remarks
 * Resolutions could be retrieved by name. The value of the name is computed as follows
 * (from highest to lowest priority):
 *
 * 1. `name` parameter for the `cache` method specified during resolution.
 *
 * 2. `name` parameter used for named registrations.
 *
 * 3. service parameter representing an object with the `name` property.
 *
 * 4. service parameter representing a string object
 *
 * @public
 */
interface ICache {
    /**
     * Retrieves resolution interface directly via name.
     * Returns service resolution instance or undefined if no resolution found
     */
    readonly instance: {
        [name: string]: any;
    };
    /**
     * Resolves an instance of a service
     * @param name - service resolution name
     * @returns - an instance of service resolution.
     * Throws {@link ResolutionError} when no resolution found
     */
    resolve<R>(name: string): R;
}//# sourceMappingURL=index.d.ts.map

/**
 * Represents fluent cascading interface for services resolution
 * @public
 */
interface IResolveWith<T> {
    /**
     * Sets arguments for service instantiation
     * @param args - optional arguments for the service instantiation
     * @returns - an instance of {@link ResolveWithArgs} interface
     */
    args(...args: any[]): ResolveWithArgs<T>;
    /**
     * Sets attempted resolution flag indicating non disruptive resolution
     * behavior (no exception is thrown when no registration found)
     * @returns - an instance of {@link ResolveWithAttempt} interface
     */
    attempt(): ResolveWithAttempt<T>;
    /**
     * Sets named resolution flag. Service will be resolved given the name provided
     * @param value - the name for the service to be resolved with
     * @returns - an instance of {@link ResolveWithName} interface
     */
    name(value: string): ResolveWithName<T>;
    /**
     * Sets dynamic dependencies for service resolution
     * @param dependencies - {@link IDynamicDependency} instance(s)
     * @returns - an instance of {@link ResolveWithName} interface
     */
    dependencies(dependencies: IDynamicDependency | IDynamicDependency[]): ResolveWithName<T>;
    /**
     * Sets cached resolution flag. When resolved, resolution will be stored in cache
     * @param name - the name to be used to retrieve resolution form cache
     * @returns - an instance of {@link ResolveWithCache} interface
     */
    cache(name?: string): ResolveWithCache<T>;
    /**
     * Executes service resolution
     * @returns - an instance of resolved service
     */
    exec(): T;
    /**
     * Asynchronously executes service resolution
     * @returns - promise, resolving with a registered instance of a service
     */
    execAsync(): Promise<T>;
}
/**
 * Represents a step within fluent cascading API resolution sequence where `args` method
 * was omitted
 * @public
 */
declare type ResolveWithArgs<T> = Omit<IResolveWith<T>, 'args'>;
/**
 * Represents a step within fluent cascading API resolution sequence where `attempt` method
 * was omitted
 * @public
 */
declare type ResolveWithAttempt<T> = Omit<ResolveWithArgs<T>, 'attempt'>;
/**
 * Represents a step within fluent cascading API resolution sequence where `name` method
 * was omitted
 * @public
 */
declare type ResolveWithName<T> = Omit<ResolveWithAttempt<T>, 'name'>;
/**
 * * Represents final step within fluent cascading API resolution sequence
 * @public
 */
declare type ResolveWithCache<T> = Pick<IResolveWith<T>, 'exec' | 'execAsync'>;
//# sourceMappingURL=resolution.d.ts.map

/**
 * Represents container interface
 *
 * @remarks
 * Containers are used to resolve prior registered services
 *
 * @public
 */
interface IContainer {
    /**
     * Represents a dictionary of cached components/resolved registrations {@link ICache}
     *
     * @remarks
     * If a service is registered with a cache option, its resolution is available
     * through the `cache` property.
     */
    cache: ICache;
    /**
     * Resolves a service with optional parameters
     * Throws {@link ResolutionError} if not registration found
     * @param service - service value registered prior resolution.
     * If service is `null` or `undefined` {@link ArgumentError} is thrown
     * @param args - optional arguments for the service instantiation
     * @returns - registered instance of a service
     */
    resolve<R>(service: {}, ...args: {}[]): R | never;
    /**
     * Asynchronously resolves a service with optional parameters
     * @param service - service value registered prior resolution
     * @param args - optional arguments for the service instantiation
     * @returns - promise, resolving with a registered instance of a service.
     * If registration not found - rejects the promise with {@link ResolutionError}
     * If `null` or `undefined` service value, rejects the promise with {@link ArgumentError}
     */
    resolveAsync<R>(service: {}, ...args: {}[]): Promise<R>;
    /**
     * Attempts to resolve a service with optional parameters.
     * @param service - service value registered prior resolution.
     * If service is `null` or `undefined` {@link ArgumentError} is thrown
     * @param args - optional arguments for the service instantiation
     * @returns - registered instance of a service.
     * If registration not found - returns undefined
     */
    tryResolve<R>(service: {}, ...args: {}[]): R | undefined;
    /**
     * Asynchronously attempts to resolve a service with optional parameters.
     * @param service - service value registered prior resolution
     * @param args - optional arguments for the service instantiation
     * @returns - promise, resolving with a registered instance of a service.
     * If registration not found - returns a promise with undefined resolution
     * If `null` or `undefined` service value rejects the promise with {@link ArgumentError}
     */
    tryResolveAsync<R>(service: {}, ...args: {}[]): Promise<R | undefined>;
    /**
     * Resolves a service with specific name and optional parameters
     * Throws {@link ResolutionError} if not registration found
     * @param service - service value registered prior resolution.
     * If service is `null` or `undefined` {@link ArgumentError} is thrown
     * @param name - a unique named used to register a service
     * If name is `null` or `undefined` {@link ArgumentError} is thrown
     * @param args - optional arguments for the service instantiation
     * @returns - registered instance of a service
     */
    resolveNamed<R>(service: {}, name: string, ...args: {}[]): R | never;
    /**
     * Asynchronously resolves a service with specific name and optional parameters
     * @param service - service value registered prior resolution
     * @param name - a unique named used to register a service
     * @param args - optional arguments for the service instantiation
     * @returns - promise, resolving with a registered instance of a service.
     * If registration not found - rejects the promise with {@link ResolutionError}
     * If `null` or `undefined` service value rejects the promise with {@link ArgumentError}
     * If `null` or `undefined` name value rejects the promise with {@link ArgumentError}
     */
    resolveNamedAsync<R>(service: {}, name: string, ...args: {}[]): Promise<R>;
    /**
     * Attempts to resolve a service with specific name and optional parameters
     * @param service - service value registered prior resolution.
     * If service is `null` or `undefined` {@link ArgumentError} is thrown
     * @param name - a unique named used to register a service
     * If name is `null` or `undefined` {@link ArgumentError} is thrown
     * @param args - optional arguments for the service instantiation
     * @returns - registered instance of a service.
     * If registration not found - returns undefined
     */
    tryResolveNamed<R>(service: {}, name: string, ...args: {}[]): R | undefined;
    /**
     * Asynchronously attempts to resolve a service with specific name and optional parameters
     * @param service - service value registered prior resolution
     * @param name - a unique named used to register a service
     * @param args- optional arguments for the service instantiation
     * @returns - promise, resolving with a registered instance of a service.
     * If registration not found - returns a promise with undefined resolution
     * If `null` or `undefined` service value rejects the promise with {@link ArgumentError}
     * If `null` or `undefined` name value rejects the promise with {@link ArgumentError}
     */
    tryResolveNamedAsync<R>(service: {}, name: string, ...args: {}[]): Promise<R | undefined>;
    /**
     * Resolves a service with dynamic dependencies
     *
     * @remarks
     * A service gets resolved with all the dependencies provided without affecting original
     * registration. All the services resolved with dynamic dependencies get transient
     * (no scope, {@link scope}) life cycle assigned regardless of initial
     * life cycle specified
     *
     * Throws {@link ResolutionError} if not registration found
     * @param service - service value registered prior resolution
     * If service is `null` or `undefined` {@link ArgumentError} is thrown
     * @param dependencies - an array of {@link IDynamicDependency} instances
     * @returns - registered instance of a service
     */
    resolveWithDependencies<R>(service: {}, dependencies: IDynamicDependency[]): R | never;
    /**
     * Asynchronously resolves a service with dynamic dependencies
     *
     * @remarks
     * A service gets resolved with all the dependencies provided without affecting original
     * registration. All the services resolved with dynamic dependencies get transient
     * (no scope, {@link scope}) life cycle assigned regardless of initial
     * life cycle specified
     *
     * @param service - service value registered prior resolution
     * @param dependencies - an array of {@link IDynamicDependency} instances
     * @returns - promise, resolving with a registered instance of a service
     * If registration not found - rejects the promise with {@link ResolutionError}
     * If `null` or `undefined` service value rejects the promise with {@link ArgumentError}
     */
    resolveWithDependenciesAsync<R>(service: {}, dependencies: IDynamicDependency[]): Promise<R>;
    /**
     * Resolves a service using fluent cascading interface
     *
     * @param service - service value registered prior resolution
     * If service is `null` or `undefined` {@link ArgumentError} is thrown
     * @returns - an instance of {@link IResolveWith} interface
     */
    resolveWith<R>(service: {}): IResolveWith<R> | never;
    /**
     * Creates nested life cycle scoped container
     * @returns - an instance of {@link IContainer} interface
     */
    createChild(): IContainer;
    /**
     * Disposes all resolved, internally owned instances registered
     * using {@link Disposer} interface
     */
    dispose(): void;
    /**
     * Asynchronously disposes all resolved, internally owned instances registered
     * using {@link Disposer} interface
     * @returns - promise
     */
    disposeAsync(): Promise<void>;
}
/**
 * Represents container builder interface
 *
 * @remarks
 * Container builders provide functionality for working with containers and registering services
 *
 * @public
 */
interface IContainerBuilder {
    /**
     * Registers a service using fluent API
     *
     * @param service - an instance of a service.
     * If service is `null` or `undefined` {@link ArgumentError} is thrown
     * @returns - an instance of {@link IRegistration} fluent API interface
     */
    register<R>(service: {}): IRegistration<R>;
    /**
     * Creates an instance of {@link IContainer} interface
     *
     * @returns - an instance of {@link IContainer} interface
     */
    build(): IContainer;
    /**
     * Copies all registrations from the container builder provided into self
     *
     * @param builder - an instance of {@link IContainerBuilder} interface
     */
    copy(builder: IContainerBuilder): void;
}//# sourceMappingURL=common.d.ts.map

/**
 * Represents an entry into service registration fluent cascading API interface
 * @public
 */
interface IDecoratorRegistration<T> {
    /**
     * Specifies initialization action during service instantiation
     * @param action - an instance of {@link Initializer} interface
     * @returns - an instance of {@link WithDecoratorRegisterInitializeBy} interface
     */
    initializeBy<K extends T>(action: Initializer<K>): WithDecoratorRegisterInitializeBy<K>;
    /**
     * Specifies lazy registration. Resolution of services is differed until results
     * are needed by other services/computations
     * @returns - an instance of {@link WithDecoratorRegisterLazy} interface
     */
    lazy(): WithDecoratorRegisterLazy<T>;
    /**
     * Specifies disposer action during resolution disposal.
     * @param action - disposer action to be called during resolution
     * disposal process ({@link Disposer})
     * @returns - an instance of {@link WithDecoratorRegisterLazy} interface
     */
    dispose<K extends T>(action: Disposer<K>): WithDecoratorRegisterLazy<K>;
    /**
     * Specifies named registration. Registration can be resolved only using the name
     * provided
     * @param name - a value that should be used during service resolution
     * If parameter value is `null` or `undefined` {@link ArgumentError} is thrown
     * @returns - an instance of {@link WithDecoratorRegisterName} interface
     */
    named(name: string): WithDecoratorRegisterName<T>;
    /**
     * Specifies resolution instance scope. Scope determines how resolved service
     * instances behave in between resolutions
     * @param scope - the value of the scope {@link ScopeType}
     * @returns - an instance of {@link WithDecoratorRegister} interface
     */
    within(scope: ScopeType): WithDecoratorRegister<T>;
    /**
     * Specifies transient scope resolution. Every resolution of service returns new instance
     * Is similar to scope.none, scope.none {@link scope} (default behavior)
     * @returns - an instance of {@link WithDecoratorRegister} interface
     */
    transient(): WithDecoratorRegister<T>;
    /**
     * Specifies singleton scope resolution. Every resolution of service returns same instance
     * Is similar to scope.hierarchy {@link scope}
     * @returns - an instance on {@link WithDecoratorRegister} interface
     */
    singleton(): WithDecoratorRegister<T>;
    /**
     * Specifies instance per container scope resolution.
     * Every resolution of a service returns the same instance per instance of a container
     * Is similar to scope.container {@link scope}
     * @returns - an instance of {@link WithDecoratorRegister} interface
     */
    instancePerContainer(): WithDecoratorRegister<T>;
    /**
     * Finalizes service registration
     */
    register(): ClassDecorator;
}
/**
 * Represents a step within fluent cascading API registration sequence where
 * registration is finalized
 * @public
 */
declare type WithDecoratorRegister<T> = Pick<IDecoratorRegistration<T>, 'register'>;
/**
 * Represents a step within fluent cascading API registration sequence where
 * `initializeBy` method was omitted
 * @public
 */
declare type WithDecoratorRegisterInitializeBy<T> = Omit<IDecoratorRegistration<T>, 'initializeBy'>;
/**
 * Represents a step within fluent cascading API registration sequence where `lazy` and
 * `dispose` methods were omitted
 * @public
 */
declare type WithDecoratorRegisterLazy<T> = Omit<WithDecoratorRegisterInitializeBy<T>, 'lazy' | 'dispose'>;
/**
 * Represents a step within fluent cascading API registration sequence where `named` method
 * was omitted
 * @public
 */
declare type WithDecoratorRegisterName<T> = Omit<WithDecoratorRegisterLazy<T>, 'named'>;
//# sourceMappingURL=registration.d.ts.map

/**
 * Represents fluent cascading interface for services resolution
 * @public
 */
interface IDecoratorResolution {
    /**
     * Sets arguments for service instantiation
     * @param args - optional arguments for the service instantiation
     * @returns - an instance of {@link WithDecoratorResolverArgs} interface
     */
    args(...value: any[]): WithDecoratorResolverArgs;
    /**
     * Sets attempted resolution flag indicating non disruptive resolution
     * behavior (no exception is thrown when no registration found)
     * @returns - an instance of {@link WithDecoratorResolverAttempt} interface
     */
    attempt(): WithDecoratorResolverAttempt;
    /**
     * Sets named resolution flag. Service will be resolved given the name provided
     * @param value - the name for the service to be resolved with
     * @returns - an instance of {@link WithDecoratorResolverName} interface
     */
    name(value: string): WithDecoratorResolverName;
    /**
     * Sets cached resolution flag. When resolved, resolution will be stored in cache
     * @param name - the name to be used to retrieve resolution form cache
     * @returns - an instance of {@link WithDecoratorResolver} interface
     */
    cache(name?: string): WithDecoratorResolver;
    /**
     * Finalizes service resolution specification
     * @returns - an instance of ParameterDecorator
     */
    resolve(): ParameterDecorator;
}
/**
 * Represents service resolution fluent API step
 * @public
 */
declare type WithDecoratorResolver = Pick<IDecoratorResolution, 'resolve'>;
/**
 * Represents service resolution fluent API step where `args` method was omitted
 * @public
 */
declare type WithDecoratorResolverArgs = Omit<IDecoratorResolution, 'args'>;
/**
 * Represents service resolution fluent API step where `attempt` method was omitted
 * @public
 */
declare type WithDecoratorResolverAttempt = Omit<WithDecoratorResolverArgs, 'attempt'>;
/**
 * Represents service resolution fluent API step where `name` method was omitted
 * @public
 */
declare type WithDecoratorResolverName = Omit<WithDecoratorResolverAttempt, 'name'>;
//# sourceMappingURL=resolution.d.ts.map

/**
 * Represents a decorator style fluent cascading API services registration /
 * resolution interface
 * @public
 */
interface IDecorator {
    /**
     * Creates an instance of {@link IContainer} interface
     *
     * @returns - an instance of {@link IContainer} interface
     */
    build(): IContainer;
    /**
     * Registers a service using fluent API
     *
     * @param service - an instance of a service.
     * If service is `null` or `undefined` {@link ArgumentError} is thrown
     * @returns - an instance of {@link IDecoratorRegistration} fluent API interface
     */
    provide<R>(service: {}): IDecoratorRegistration<R>;
    /**
     * Sets registration to be marked as self registration
     *
     * @remarks
     * Registration is provided as a type itself (participates both in registration and
     * resolution). It is intended for construct-able types only. Uses new
     * operator for type construction
     *
     * @returns - an instance of {@link IDecoratorRegistration} interface
     */
    provideSelf<R>(): IDecoratorRegistration<R>;
    /**
     * Specifies entry point into decorator fluent cascading API for
     * service resolution
     * @param service - component registration service
     * @returns - an instance of {@link IDecoratorResolution} interface
     */
    by(service?: {}): IDecoratorResolution;
    /**
     * Specifies a value to be used for parameter resolution. It is used as is
     * and no resolution computation is preformed
     * @param value - resolution value
     * @returns - ParameterDecorator instance
     */
    resolveValue(value: {} | Function): ParameterDecorator;
    /**
     * Provides non decorator compatible registration interface
     * @param service - component registration service
     * @returns an instance on {@link IRegistration} interface
     */
    register<R>(service: {}): IRegistration<R>;
    /**
     * Imports registrations from non decorator container builder
     * @param builder - an instance of {@link IContainerBuilder} interface
     */
    import(builder: IContainerBuilder): void;
}
//# sourceMappingURL=common.d.ts.map

/**
 * Represents a substitute parameters specification interface
 * @public
 */
interface ICallInfo {
    /**
     * Original subject
     */
    source: {};
    /**
     * Original subject member name
     */
    name: string;
    /**
     * Substitute invocation parameters
     */
    args: any[];
    /**
     * Invokes original subject member
     * @param args - optional array of invocation parameters
     * @returns - for methods/getters/fields substitution execution result, otherwise void
     */
    invoke: (args?: any | any[]) => any | void;
    /**
     * Original subject member type {@link CallInfoType}
     */
    type: CallInfoType;
    /**
     * Invokes original subject getter. If original subject member is not a getter nor a field,
     * this member is undefined
     */
    get?: () => any;
    /**
     * Invokes original subject setter. If original subject member is not a setter nor a field,
     * this member is undefined
     */
    set?: ({}: {}) => void;
    /**
     * Invokes next substitution in a chain. If only one substitution is present in the chain,
     * this member is undefined
     */
    next?: (result?: any) => any;
    /**
     * Represents a result of a previous substitution in a chain, if it exist
     */
    result?: any;
}
/**
 * Represents substitute information encapsulation interface
 *
 * @remarks
 * Multiple substitutions for the same member form an execution chain.
 * Every predecessor in the chain can pass a result of the execution to its successor.
 * Every successor in the chain can receive a result of predecessor execution.
 * If substitution does not invoke its successor, execution chain stops
 *
 * @public
 */
interface ISubstituteInfo {
    /**
     * Original subject member name. If omitted, substitution is applied to all members
     */
    method?: string;
    /**
     * Original subject member type. Used to specify partial substitution of
     * getters and setters. If omitted, original subject member type is used {@link CallInfoType}
     */
    type?: CallInfoType;
    /**
     * Specifies substitute action/behavior (lambda expression executed during substitution call)
     *
     * @remarks
     * To execute wrapper in the context of the substitute proxy as opposite to original
     * subject, provide a function instead of a lambda expression. This will give you
     * access to `this` within substituted members
     *
     * @param callInfo - an instance of {@link ICallInfo} interface
     * @returns - for methods/getters/fields substitution execution result, otherwise void
     */
    wrapper: (callInfo: ICallInfo) => any | void;
}//# sourceMappingURL=common.d.ts.map

/**
 * Represents fluent cascading API interface for substitutes specification
 * @public
 */
interface IWithSubstituteResult {
    /**
     * Adds a new substitute to the substitute chain
     * @param substitute - an instance of {@link ISubstituteInfo} interface
     * @returns self representative to be used for future substitutes
     * specification
     */
    withSubstitute: (substitute: ISubstituteInfo) => IWithSubstituteResult;
    /**
     * Creates intercepted instance representative
     * @param subject - an instance of a type
     * @returns - a decorated instance
     */
    interceptInstance: <R extends Object>(subject: R) => R;
    /**
     * Creates intercepted prototype representative
     * @param subject - function/class prototype (prototypical/constructible function)
     * @returns - a decorated prototype.
     */
    interceptPrototype: <R extends Function>(subject: R) => R;
    /**
     * Create a subject with members substituted by given parameters.
     * For the instances of an object - interceptInstance method is used,
     * for prototypical/constructible functions - interceptPrototype is used
     *
     * @param subject - prototypical/constructable function or an instance of an abject
     * @returns - decorated subject
     */
    intercept: <R extends (Function | Object)>(subject: R) => R;
}
/**
 * Represents interceptor specification interface
 *
 * @remarks
 * Interceptors are used to intercept calls to methods/getters/setters/fields of provided component
 *
 * @public
 */
interface IInterceptor {
    /**
     * Creates intercepted prototype representative
     * @param subject - function/class prototype (prototypical/constructible function)
     * @param substitutes - instance(s) of {@link ISubstituteInfo} interface.
     * If substitutes are applied to the same member, they are chained
     * @returns - a decorated prototype. If no substitutes provided
     * returns same input subject parameter
     */
    interceptPrototype<R extends Function>(subject: R, substitutes?: ISubstituteInfo | ISubstituteInfo[]): R;
    /**
     * Creates intercepted instance representative
     * @param subject - an instance of a type
     * @param substitutes - instance(s) of {@link ISubstituteInfo} interface.
     * If substitutes are applied to the same member, they are chained
     * @returns - a decorated instance. If no substitutes provided
     * returns same input subject parameter
     */
    interceptInstance<R extends Object>(subject: R, substitutes?: ISubstituteInfo | ISubstituteInfo[]): R;
    /**
     * Create a subject with members substituted by given parameters.
     * For the instances of an object - interceptInstance method is used,
     * for prototypical/constructible functions - interceptPrototype is used
     *
     * @param subject - prototypical/constructable function or an instance of an abject
     * @param substitutes - instance(s) of {@link ISubstituteInfo} interface
     * @returns - decorated subject. If no substitutes provided returns same input
     * subject parameter
     */
    intercept<R extends Function | Object>(subject: R, substitutes?: ISubstituteInfo | ISubstituteInfo[]): R;
    /**
     * Represents an entry point in fluent cascading API substitutes specification
     * @param substitute - an instance of {@link ISubstituteInfo} interface
     * @returns - an instance of {@link IWithSubstituteResult} interface
     */
    withSubstitute(substitute: ISubstituteInfo): IWithSubstituteResult;
}
//# sourceMappingURL=interceptor.d.ts.map

/**
 * Represents general purpose application error
 * @public
 */
declare class ApplicationError extends Error {
    /**
     * @internal
     */
    _data?: {};
    get data(): {} | undefined;
    /**
     * Constructor
     * @param params - optional additional information about the error
     * @returns - an instance of ApplicationError
     */
    constructor(params?: {
        message?: string;
        data?: {};
    });
}
//# sourceMappingURL=application-error.d.ts.map

/**
 * Represents argument error. It is thrown when the argument value does not
 * comply with the method specification
 * @public
 */
declare class ArgumentError extends ApplicationError {
    /**
     * @internal
     */
    _argumentName: string;
    get argumentName(): string;
    /**
     * Constructor
     * @param argumentName - argument name
     * @param params - optional additional information about the error
     * @returns - an instance of ArgumentError
     */
    constructor(argumentName: string, params?: {
        message?: string;
        data?: {};
    });
}
//# sourceMappingURL=argument-error.d.ts.map

/**
 * Represents circular dependency error. It is thrown when two or more service
 * resolutions depend on each other directly on indirectly
 * @public
 */
declare class CircularDependencyError extends ApplicationError {
    /**
     * @internal
     */
    _serviceName: string;
    /**
     * The name of the service that circularly resolves itself
     */
    get serviceName(): string;
    /**
     * Constructor
     * @param serviceName - service name
     * @param data - optional additional information about the error
     * @returns - an instance of CircularDependencyError
     */
    constructor(serviceName: string, data?: {});
}
//# sourceMappingURL=circular-dependency-error.d.ts.map

/**
 * Represents decorator error. It is thrown when an exception happens
 * during service parts decoration (class, parameter, etc...)
 * @public
 */
declare class DecoratorError extends ApplicationError {
    /**
     * Constructor
     * @param params - optional additional information about the error
     * @returns - an instance of DecoratorError
     */
    constructor(params?: {
        message?: string;
        data?: {};
    });
}
//# sourceMappingURL=decorator-error.d.ts.map

/**
 * Represents proxy error. It is thrown when an exception happens
 * during proxy - interceptor construction
 * @public
 */
declare class ProxyError extends ApplicationError {
    /**
     * Constructor
     * @param params - optional additional information about the error
     * @returns - an instance of ProxyError
     */
    constructor(params?: {
        message?: string;
        data?: {};
    });
}
//# sourceMappingURL=proxy-error.d.ts.map

/**
 * Represents resolution error. It is thrown when an error happens
 * during service resolution / instantiation
 * @public
 */
declare class ResolutionError extends ApplicationError {
    /**
     * @internal
     */
    _innerError?: Error;
    /**
     * Internal error
     */
    get innerError(): Error | undefined;
    /**
     * Constructor
     * @param params - optional additional information about the error
     * @returns - an instance of ResolutionError
     */
    constructor(params?: {
        message?: string;
        data?: {};
        error?: Error;
    });
}
//# sourceMappingURL=resolution-error.d.ts.map

/**
 * Creates an instance on {@link IContainerBuilder} interface
 * @public
 * @returns - an instance on {@link IContainerBuilder} interface
 */
declare const builder: () => IContainerBuilder;
/**
 * Creates an instance on {@link IDecorator} interface
 * @public
 * @returns - an instance on {@link IDecorator} interface
 */
declare const decorator: () => IDecorator;
/**
 * Creates an instance on {@link IInterceptor} interface
 * @public
 * @returns - an instance on {@link IInterceptor} interface
 */
declare const interceptor: () => IInterceptor;//# sourceMappingURL=index.d.ts.map

export { ApplicationError, ArgumentError, CallInfoType, CircularDependencyError, DecoratorError, Disposer, Factory, ICache, ICallInfo, IContainer, IContainerBuilder, IDecorator, IDecoratorRegistration, IDecoratorResolution, IDynamicDependency, IInterceptor, IName, IRegisterWithAs, IRegistration, IResolveWith, ISubstituteInfo, IWithSubstituteResult, Initializer, ProxyError, RegisterWithInitializeBy, RegisterWithLazy, RegisterWithName, ResolutionError, ResolveWithArgs, ResolveWithAttempt, ResolveWithCache, ResolveWithName, ScopeType, WithDecoratorRegister, WithDecoratorRegisterInitializeBy, WithDecoratorRegisterLazy, WithDecoratorRegisterName, WithDecoratorResolver, WithDecoratorResolverArgs, WithDecoratorResolverAttempt, WithDecoratorResolverName, builder, callInfo, decorator, interceptor, scope };
