import 'reflect-metadata';
import { HashTable } from 'hashes';

class DisposableStorage {
    constructor() {
        this._disposables = [];
    }
    add(obj, disposer) {
        const item = this.createDisposableItem(obj, disposer);
        this._disposables.push(item);
    }
    disposeItems() {
        while (this._disposables.length > 0) {
            const item = this._disposables.pop();
            const obj = item.weakReference;
            item.disposer(obj);
        }
    }
    createDisposableItem(obj, disposer) {
        return {
            weakReference: obj,
            disposer
        };
    }
}

class InternalStorage {
    constructor() {
        this._collection = new HashTable();
    }
    add(key, value) {
        this._collection.add(key, value, true);
    }
    get(key) {
        return this._collection.get(key).value;
    }
    tryGet(key) {
        return this.contains(key) ? this._collection.get(key).value : undefined;
    }
    register(key, defaultValue) {
        if (this.contains(key)) {
            return this._collection.get(key).value;
        }
        const entry = defaultValue();
        this.add(key, entry);
        return entry;
    }
    contains(key) {
        return this._collection.contains(key);
    }
    clear() {
        this._collection.clear();
    }
}

const registrationType = Object.freeze({
    factory: 1,
    factoryType: 2,
    factoryValue: 3
});

/**
 * Represents ownerships type
 */
const owner = Object.freeze({
    /**
     * Container ownership
     */
    container: 1,
    /**
     * External ownership
     */
    externals: 2
});

/**
 * Represents resolution life-cycle type definition
 * @public
 */
const scope = Object.freeze({
    /**
     * Transient life-cycle, one instance per resolution
     */
    none: 1,
    /**
     * Container life-cycle, one instance per resoluton per container instance
     */
    container: 2,
    /**
     * Hierarchy life-cycle, same instance per all resolutions
     */
    hierarchy: 3
});

/**
 * Represents original substituted subject member type
 * @public
 */
const callInfo = Object.freeze({
    /**
     * Indicates method member type substitution
     */
    method: 1,
    /**
     * Indicates getter member type substitution
     */
    getter: 2,
    /**
     * Indicates setter member type substitution
     */
    setter: 3,
    /**
     * Indicates substitution for both getter and setter member types
     */
    getterSetter: 4,
    /**
     * Indicates substitution of any member type
     */
    any: 5,
    /**
     * Indicates field substitution
     */
    field: 6
});

const defaults = {
    get scope() {
        return scope.none;
    },
    get owner() {
        return owner.container;
    }
};

function setPrototypeOf(instance, prototype) {
    if (Object.setPrototypeOf) {
        Object.setPrototypeOf(instance, prototype);
    }
}

/**
 * Represents general purpose application error
 * @public
 */
class ApplicationError extends Error {
    /**
     * Constructor
     * @param params - optional additional information about the error
     * @returns - an instance of ApplicationError
     */
    constructor(params) {
        super(params ? params.message : undefined);
        setPrototypeOf(this, ApplicationError.prototype);
        this.name = 'Application Error';
        this._data = params ? params.data : undefined;
    }
    get data() {
        return this._data;
    }
}

/**
 * Represents argument error. It is thrown when the argument value does not
 * comply with the method specification
 * @public
 */
class ArgumentError extends ApplicationError {
    /**
     * Constructor
     * @param argumentName - argument name
     * @param params - optional additional information about the error
     * @returns - an instance of ArgumentError
     */
    constructor(argumentName, params) {
        super(params);
        setPrototypeOf(this, ArgumentError.prototype);
        this.name = 'Argument Error';
        this._argumentName = argumentName;
    }
    get argumentName() {
        return this._argumentName;
    }
}

/**
 * Represents circular dependency error. It is thrown when two or more service
 * resolutions depend on each other directly on indirectly
 * @public
 */
class CircularDependencyError extends ApplicationError {
    /**
     * Constructor
     * @param serviceName - service name
     * @param data - optional additional information about the error
     * @returns - an instance of CircularDependencyError
     */
    constructor(serviceName, data) {
        super({ message: `Circular dependency for service: ${serviceName}`, data });
        setPrototypeOf(this, CircularDependencyError.prototype);
        this.name = 'Circular Dependency Error';
        this._serviceName = serviceName;
    }
    /**
     * The name of the service that circularly resolves itself
     */
    get serviceName() {
        return this._serviceName;
    }
}

/**
 * Represents decorator error. It is thrown when an exception happens
 * during service parts decoration (class, parameter, etc...)
 * @public
 */
class DecoratorError extends ApplicationError {
    /**
     * Constructor
     * @param params - optional additional information about the error
     * @returns - an instance of DecoratorError
     */
    constructor(params) {
        super(params);
        setPrototypeOf(this, DecoratorError.prototype);
        this.name = 'Decorator Error';
    }
}

/**
 * Represents proxy error. It is thrown when an exception happens
 * during proxy - interceptor construction
 * @public
 */
class ProxyError extends ApplicationError {
    /**
     * Constructor
     * @param params - optional additional information about the error
     * @returns - an instance of ProxyError
     */
    constructor(params) {
        super(params);
        setPrototypeOf(this, ProxyError.prototype);
        this.name = 'Proxy Error';
    }
}

/**
 * Represents resolution error. It is thrown when an error happens
 * during service resolution / instantiation
 * @public
 */
class ResolutionError extends ApplicationError {
    /**
     * Constructor
     * @param params - optional additional information about the error
     * @returns - an instance of ResolutionError
     */
    constructor(params) {
        super(params);
        setPrototypeOf(this, ResolutionError.prototype);
        this.name = 'Resolution Error';
        this._innerError = params ? params.error : undefined;
    }
    /**
     * Internal error
     */
    get innerError() {
        return this._innerError;
    }
}

function getMetadata(type) {
    return Reflect.getMetadata('design:paramtypes', type) || [];
}
function construct(constructor, args) {
    return args && args.length ? new constructor(...args) : new constructor();
}
function isArray(value) {
    return Array.isArray(value);
}
function isFunction(f) {
    return f instanceof Function;
}
function isPrototype(f) {
    return toString.call(f) === '[object Function]';
}
function isObject(o) {
    return o === Object(o);
}
function getPropertyDescriptor(object, key) {
    let descriptor;
    do {
        descriptor = Object.getOwnPropertyDescriptor(object, key);
    } while (!descriptor && (object = Object.getPrototypeOf(object)));
    return descriptor;
}
function getAllPropertyNames(obj) {
    const props = [];
    do {
        Object.getOwnPropertyNames(obj).forEach((prop) => {
            if (props.indexOf(prop) === -1) {
                props.push(prop);
            }
        });
    } while (obj = Object.getPrototypeOf(obj));
    return props;
}

const factoryValueKey = 'factoryValue';
function checkNullArgument(value, argument, message) {
    if (value == null || value === undefined) {
        throw new ArgumentError(argument, { message });
    }
}
function checkDependency(dependency) {
    if ((dependency.factory && dependency.factoryType) ||
        (dependency.factory && factoryValueKey in dependency) ||
        (dependency.factoryType && factoryValueKey in dependency)) {
        throw new ApplicationError({ message: 'Unknown registration type' });
    }
}
function uuid() {
    const replace = (c) => {
        // tslint:disable-next-line: one-variable-per-declaration
        const r = Math.random() * 16 | 0, v = c === 'x' ? r : (r & 0x3 | 0x8);
        return v.toString(16);
    };
    return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, replace);
}

class ImmutableArray {
    constructor(data) {
        checkNullArgument(data, 'data');
        if (!isArray(data)) {
            throw new ArgumentError('data', { message: 'should represent an array' });
        }
        this._data = this.initialize(data);
    }
    get value() {
        return this._data.slice(0);
    }
    static createImmutable(array) {
        return new ImmutableArray(array);
    }
    initialize(data) {
        return data.map(item => {
            const type = typeof (item);
            return type === 'object' ||
                type === 'function' ? Object.freeze(item) : item;
        });
    }
}

class RegistrationBase {
    constructor(_service, id) {
        this._service = _service;
        this._owner = owner.externals;
        this._args = [];
        this._params = [];
        this._dependenciesValue = [];
        this._isLazy = false;
        this.args = [];
        this.params = [];
        this._id = id || uuid();
    }
    get name() {
        return this._name;
    }
    set name(value) {
        this._name = value;
    }
    get service() {
        return this._service;
    }
    get scope() {
        return this._scope;
    }
    set scope(value) {
        this._scope = value;
    }
    get owner() {
        return this._owner;
    }
    set owner(value) {
        this._owner = value;
    }
    get initializer() {
        return this._initializer;
    }
    set initializer(value) {
        this._initializer = value;
    }
    get disposer() {
        return this._disposer;
    }
    set disposer(value) {
        this._disposer = value;
    }
    get args() {
        return this._args;
    }
    set args(value) {
        this._args = value;
    }
    get params() {
        return this._params;
    }
    set params(value) {
        this._params = value;
    }
    get container() {
        return this._container;
    }
    set container(value) {
        this._container = value;
    }
    get instance() {
        return this._instance;
    }
    set instance(value) {
        this._instance = value;
    }
    get registrationType() {
        if (!this._registrationType) {
            throw new ApplicationError({
                message: 'Unknown registration type',
                data: this._registrationType
            });
        }
        return this._registrationType;
    }
    get dependenciesValue() {
        return this._dependenciesValue;
    }
    set dependenciesValue(value) {
        this._dependenciesValue = isArray(value) ? value : [];
    }
    get factory() {
        return this._factory;
    }
    set factory(value) {
        this._factory = value;
        this._registrationType = registrationType.factory;
    }
    get factoryType() {
        return this._factoryType;
    }
    set factoryType(value) {
        this._factoryType = value;
        this._registrationType = registrationType.factoryType;
    }
    get factoryValue() {
        return this._factoryValue;
    }
    set factoryValue(value) {
        this._factoryValue = value;
        this._registrationType = registrationType.factoryValue;
    }
    get id() {
        return this._id;
    }
    get isLazy() {
        return this._isLazy;
    }
    set isLazy(value) {
        this._isLazy = value;
    }
    cloneFor(container) {
        const result = this.clone();
        result.container = container;
        return result;
    }
    clone() {
        const result = new RegistrationBase(this._service, this._id);
        result._factory = this._factory;
        result._factoryType = this._factoryType;
        result._factoryValue = this._factoryValue;
        result._registrationType = this._registrationType;
        result.owner = this._owner;
        result.scope = this._scope;
        result.initializer = this._initializer;
        result.isLazy = this._isLazy;
        result.params = this._params;
        result.dependenciesValue = this._dependenciesValue;
        return result;
    }
    copyDependency(dependency) {
        this.name = dependency.named;
        this.initializer = dependency.initializer;
        if (dependency.factory) {
            this.factory = dependency.factory;
        }
        if (dependency.factoryType) {
            this.factoryType = dependency.factoryType;
        }
        if (factoryValueKey in dependency) {
            this.factoryValue = dependency.factoryValue;
        }
    }
    checkRegistrationType() {
        /// throws exception when no type
        this.registrationType;
    }
}

class Registration {
    constructor(_base) {
        this._base = _base;
        this.initializeBy = this.initializeBy.bind(this);
        this.dispose = this.dispose.bind(this);
        this.named = this.named.bind(this);
        this.name = this.name.bind(this);
        this.within = this.within.bind(this);
        this.ownedBy = this.ownedBy.bind(this);
        this.transient = this.transient.bind(this);
        this.singleton = this.singleton.bind(this);
        this.instancePerContainer = this.instancePerContainer.bind(this);
        this.lazy = this.lazy.bind(this);
    }
    as(factory) {
        checkNullArgument(factory, 'factory');
        this._base.factory = factory;
        return {
            initializeBy: this.initializeBy,
            lazy: this.lazy,
            dispose: this.dispose,
            named: this.named,
            within: this.within,
            transient: this.transient,
            singleton: this.singleton,
            instancePerContainer: this.instancePerContainer
        };
    }
    asType(type, ...params) {
        checkNullArgument(type, 'type');
        this._base.factoryType = type;
        this._base.params = params;
        return {
            initializeBy: this.initializeBy,
            lazy: this.lazy,
            dispose: this.dispose,
            named: this.named,
            within: this.within,
            transient: this.transient,
            singleton: this.singleton,
            instancePerContainer: this.instancePerContainer
        };
    }
    asSelf(...params) {
        return this.asType(this._base.service, ...params);
    }
    asValue(value) {
        checkNullArgument(value, 'value');
        this._base.factoryValue = value;
        this._base.owner = owner.externals;
        this._base.scope = scope.none;
        return {
            named: this.name
        };
    }
    name(value) {
        this.named(value);
    }
    named(value) {
        checkNullArgument(value, 'value');
        this._base.name = value;
        return {
            within: this.within,
            transient: this.transient,
            singleton: this.singleton,
            instancePerContainer: this.instancePerContainer
        };
    }
    within(scope) {
        checkNullArgument(scope, 'scope');
        this._base.scope = scope;
    }
    transient() {
        return this.within(scope.none);
    }
    singleton() {
        return this.within(scope.hierarchy);
    }
    instancePerContainer() {
        return this.within(scope.container);
    }
    ownedBy(owner) {
        checkNullArgument(owner, 'owner');
        this._base.owner = owner;
    }
    initializeBy(action) {
        checkNullArgument(action, 'action');
        this._base.initializer = action;
        return {
            lazy: this.lazy,
            dispose: this.dispose,
            named: this.named,
            within: this.within,
            transient: this.transient,
            singleton: this.singleton,
            instancePerContainer: this.instancePerContainer
        };
    }
    dispose(action) {
        checkNullArgument(action, 'action');
        this._base.disposer = action;
        this.ownedBy(owner.container);
        return {
            named: this.named,
            within: this.within,
            transient: this.transient,
            singleton: this.singleton,
            instancePerContainer: this.instancePerContainer
        };
    }
    lazy() {
        this._base.isLazy = true;
        return {
            named: this.named,
            within: this.within,
            transient: this.transient,
            singleton: this.singleton,
            instancePerContainer: this.instancePerContainer
        };
    }
}

const empty = {
    typeFactoryBucket() {
        return {
            factory: {
                noName: undefined,
                names: {}
            },
            type: registrationType.factoryType
        };
    },
    valueFactoryBucket() {
        return {
            factory: {
                noName: undefined,
                names: {}
            },
            type: registrationType.factoryValue
        };
    },
    factoryBucket() {
        return {
            factory: {
                noName: {},
                names: {}
            },
            type: registrationType.factory
        };
    }
};

class RegistrationStorage {
    constructor(storageService) {
        this._addStrategy = [];
        this._getStrategy = [];
        this._internalStorage = storageService.create();
        this._addStrategy[registrationType.factoryType] = this.addForFactoryType.bind(this);
        this._addStrategy[registrationType.factory] = this.addForFactory.bind(this);
        this._addStrategy[registrationType.factoryValue] = this.addForFactoryValue.bind(this);
        this._getStrategy[registrationType.factoryType] =
            this.getForFactoryType.bind(this);
        this._getStrategy[registrationType.factory] =
            this.getForFactory.bind(this);
        this._getStrategy[registrationType.factoryValue] =
            this.getForFactoryValue.bind(this);
    }
    addEntry(registration) {
        const strategy = this._addStrategy[registration.registrationType];
        strategy(registration);
    }
    getEntry(registration) {
        const storage = this._internalStorage.tryGet(registration.service);
        if (!storage)
            return undefined;
        return this._getStrategy[storage.type](registration, storage);
    }
    clear() {
        this._internalStorage.clear();
    }
    addForFactoryType(registration) {
        const storage = this._internalStorage.register(registration.service, empty.typeFactoryBucket);
        if (!registration.name) {
            storage.factory.noName = registration;
        }
        else {
            storage.factory.names[registration.name] = registration;
        }
    }
    addForFactory(registration) {
        const storage = this._internalStorage.register(registration.service, empty.factoryBucket);
        const argsCount = this.getArgumentsCount(registration);
        if (!registration.name) {
            storage.factory.noName[argsCount] = registration;
        }
        else {
            const bucket = storage.factory.names[registration.name] || {};
            bucket[argsCount] = registration;
            storage.factory.names[registration.name] = bucket;
        }
    }
    addForFactoryValue(registration) {
        const storage = this._internalStorage.register(registration.service, empty.valueFactoryBucket);
        storage.type = registrationType.factoryValue;
        if (!registration.name) {
            storage.factory.noName = registration;
        }
        else {
            storage.factory.names[registration.name] = registration;
        }
    }
    getForFactoryType(registration, storage) {
        return !registration.name ? storage.factory.noName :
            storage.factory.names[registration.name];
    }
    getForFactory(registration, storage) {
        const argsCount = this.getArgumentsCount(registration);
        const name = registration.name ? storage.factory.names[registration.name] : undefined;
        return registration.name ?
            (name ? name[argsCount] : undefined) :
            storage.factory.noName[argsCount];
    }
    getForFactoryValue(registration, storage) {
        return !registration.name ? storage.factory.noName :
            storage.factory.names[registration.name];
    }
    getArgumentsCount(registration) {
        return registration.factory ?
            this.getFactoryArgsCount(registration.factory) :
            registration.args.length;
    }
    getFactoryArgsCount(factory) {
        const paramsCount = factory.length;
        return Math.max(paramsCount - 1, 0);
    }
}

class ResolutionCache {
    constructor() {
        this._instance = {};
        this._storage = new InternalStorage();
    }
    get instance() {
        return this._instance;
    }
    resolve(name) {
        if (!this._storage.contains(name)) {
            throw new ResolutionError({ message: 'Could not resolve service', data: name });
        }
        return this._storage.get(name);
    }
    add(name, value) {
        this._storage.add(name, value);
        Object.defineProperty(this.instance, name, {
            enumerable: true,
            configurable: false,
            get: () => {
                return this._storage.get(name);
            }
        });
    }
    clear() {
        this._storage.clear();
        this._instance = {};
    }
}

class ContainerBuilder {
    constructor(_registrationBaseService, _instanceRegistrationService, _internalContainerService, _containerService) {
        this._registrationBaseService = _registrationBaseService;
        this._instanceRegistrationService = _instanceRegistrationService;
        this._internalContainerService = _internalContainerService;
        this._containerService = _containerService;
        this._registrations = [];
    }
    register(service) {
        checkNullArgument(service, 'service');
        const registrationBase = this._registrationBaseService.create(service);
        const registration = this._instanceRegistrationService.create(registrationBase);
        setDefaults(registrationBase);
        this._registrations.push(registrationBase);
        return registration;
    }
    build() {
        const registrations = this._registrations.slice(0);
        const internalContainer = this._internalContainerService.create();
        const container = this._containerService.create(internalContainer);
        internalContainer.add(registrations);
        return container;
    }
    copy(builder) {
        checkNullArgument(builder, 'builder');
        builder
            ._registrations
            .forEach((item) => {
            const clone = item.clone();
            const index = this._registrations
                .findIndex((registration) => registration.service === item.service);
            if (index < 0) {
                this._registrations.push(clone);
            }
            else {
                this._registrations[index] = clone;
            }
        });
    }
}
function setDefaults(registration) {
    registration.scope = defaults.scope;
    registration.owner = defaults.owner;
}

const container = (_container) => {
    const result = {
        get cache() {
            return _container.cache;
        },
        createChild() {
            return container(_container.createChild());
        },
        dispose() {
            _container.dispose();
        },
        async disposeAsync() {
            return new Promise(resolve => {
                _container.dispose();
                resolve();
            });
        },
        resolve(service, ...args) {
            checkNullArgument(service, 'service');
            if (!args.length) {
                return _container.resolve(service);
            }
            return _container.resolve(service, ...args);
        },
        async resolveAsync(service, ...args) {
            return new Promise(resolve => {
                resolve(result.resolve(service, ...args));
            });
        },
        tryResolve(service, ...args) {
            checkNullArgument(service, 'service');
            if (!args.length) {
                return _container.tryResolve(service) || undefined;
            }
            return _container.tryResolve(service, ...args);
        },
        async tryResolveAsync(service, ...args) {
            return new Promise(resolve => {
                resolve(result.tryResolve(service, ...args) || undefined);
            });
        },
        resolveNamed(service, name, ...args) {
            checkNullArgument(service, 'service');
            checkNullArgument(name, 'name');
            if (!args.length) {
                return _container.resolveNamed(service, name);
            }
            return _container.resolveNamed(service, name, ...args);
        },
        async resolveNamedAsync(service, name, ...args) {
            return new Promise(resolve => {
                resolve(result.resolveNamed(service, name, ...args));
            });
        },
        tryResolveNamed(service, name, ...args) {
            checkNullArgument(service, 'service');
            checkNullArgument(name, 'name');
            const result = args.length <= 0 ?
                _container.tryResolveNamed(service, name) :
                _container.tryResolveNamed(service, name, ...args);
            return result || undefined;
        },
        async tryResolveNamedAsync(service, name, ...args) {
            return new Promise(resolve => {
                resolve(result.tryResolveNamed(service, name, ...args) || undefined);
            });
        },
        resolveWithDependencies(service, dependencies) {
            checkNullArgument(service, 'service');
            if (!dependencies || dependencies.length <= 0) {
                throw new ResolutionError({ message: 'No dependencies provided' });
            }
            dependencies.forEach(checkDependency);
            return _container.resolveWithDependencies(service, dependencies);
        },
        async resolveWithDependenciesAsync(service, dependencies) {
            return new Promise(resolve => {
                resolve(result.resolveWithDependencies(service, dependencies));
            });
        },
        resolveWith(service) {
            checkNullArgument(service, 'service');
            return _container.resolveWith(service);
        }
    };
    return result;
};

class Api {
    constructor(_container) {
        this._container = _container;
        this._service = undefined;
        this._cache = {
            use: false,
            name: undefined
        };
        this._dependencies = [];
        this._attempt = false;
        this._args = [];
        this.args = this.args.bind(this);
        this.attempt = this.attempt.bind(this);
        this.name = this.name.bind(this);
        this.dependencies = this.dependencies.bind(this);
        this.cache = this.cache.bind(this);
        this.exec = this.exec.bind(this);
        this.execAsync = this.execAsync.bind(this);
    }
    get serviceValue() {
        if (this._service === null || this._service === undefined) {
            throw new ApplicationError({ message: 'Service was not initialized' });
        }
        return this._service;
    }
    get nameValue() {
        return this._name;
    }
    get cacheValue() {
        return this._cache;
    }
    get dependenciesValue() {
        return this._dependencies;
    }
    get isDependenciesResolvable() {
        return this._dependencies && this._dependencies.length > 0;
    }
    get attemptValue() {
        return this._attempt;
    }
    get throwResolveError() {
        return !this.attemptValue;
    }
    get argsValue() {
        return this._args;
    }
    service(value) {
        checkNullArgument(value, 'value');
        this._service = value;
        return {
            args: this.args,
            attempt: this.attempt,
            name: this.name,
            dependencies: this.dependencies,
            cache: this.cache,
            exec: this.exec,
            execAsync: this.execAsync
        };
    }
    args(...args) {
        this._args = args;
        return {
            attempt: this.attempt,
            name: this.name,
            dependencies: this.dependencies,
            cache: this.cache,
            exec: this.exec,
            execAsync: this.execAsync
        };
    }
    attempt() {
        this._attempt = true;
        return {
            name: this.name,
            dependencies: this.dependencies,
            cache: this.cache,
            exec: this.exec,
            execAsync: this.execAsync
        };
    }
    name(value) {
        checkNullArgument(value, 'value');
        this._name = value;
        return {
            dependencies: this.dependencies,
            cache: this.cache,
            exec: this.exec,
            execAsync: this.execAsync
        };
    }
    dependencies(data) {
        checkNullArgument(data, 'data');
        if (isArray(data)) {
            data
                .forEach(item => checkDependency(item));
            this._dependencies.push.apply(this._dependencies, data);
        }
        else {
            checkDependency(data);
            this._dependencies.push(data);
        }
        return {
            dependencies: this.dependencies,
            cache: this.cache,
            exec: this.exec,
            execAsync: this.execAsync
        };
    }
    cache(name) {
        this._cache.use = true;
        this._cache.name = name;
        return {
            exec: this.exec,
            execAsync: this.execAsync
        };
    }
    exec() {
        return this._container.execute(this);
    }
    async execAsync() {
        return new Promise(resolve => { resolve(this.exec()); });
    }
}

class InternalContainer {
    constructor(_registrationStorageService, _disposableStorageService, _registrationBaseService, _containerApiService, _invokerService, _resolutionCacheService, _resolutionDetails) {
        this._registrationStorageService = _registrationStorageService;
        this._disposableStorageService = _disposableStorageService;
        this._registrationBaseService = _registrationBaseService;
        this._containerApiService = _containerApiService;
        this._invokerService = _invokerService;
        this._resolutionCacheService = _resolutionCacheService;
        this._resolutionDetails = _resolutionDetails;
        this._children = [];
        this._dependencyScope = scope.none;
        this._dependencyOwner = owner.externals;
        this._collection = this._registrationStorageService.create();
        this._disposableStorage = this._disposableStorageService.create();
        this._invoker = this._invokerService.create(this, _resolutionDetails);
        this._pendingResolutions = {};
        this.registerImpl = this.registerImpl.bind(this);
        this._resolutionCache = _resolutionCacheService.create();
        this._cache = this.createCache(this._resolutionCache);
    }
    get cache() {
        return this._cache;
    }
    add(registrations) {
        registrations.forEach(this.registerImpl);
    }
    createChild() {
        const child = new InternalContainer(this._registrationStorageService, this._disposableStorageService, this._registrationBaseService, this._containerApiService, this._invokerService, this._resolutionCacheService, this._resolutionDetails);
        child._parent = this;
        this._children.push(child);
        return child;
    }
    dispose() {
        this._disposableStorage.disposeItems();
        while (this._children.length > 0) {
            const item = this._children.pop();
            item.dispose();
        }
        this._collection.clear();
        this._resolutionCache.clear();
    }
    disposeAsync() {
        throw new Error('Not implemented');
    }
    resolve(service, ...args) {
        const registration = this.createRegistration(service);
        registration.args = args;
        return this.resolveBase(registration, true);
    }
    resolveAsync(_service, ..._args) {
        throw new Error('Not implemented');
    }
    tryResolve(service, ...args) {
        const registration = this.createRegistration(service);
        registration.args = args;
        return this.resolveBase(registration, false);
    }
    tryResolveAsync(_service, ..._args) {
        throw new Error('Not implemented');
    }
    resolveNamed(service, name, ...args) {
        const registration = this.createRegistration(service);
        registration.name = name;
        registration.args = args;
        return this.resolveBase(registration, true);
    }
    resolveNamedAsync(_service, _name, ..._args) {
        throw new Error('Not implemented');
    }
    tryResolveNamed(service, name, ...args) {
        const registration = this.createRegistration(service);
        registration.name = name;
        registration.args = args;
        return this.resolveBase(registration, false);
    }
    tryResolveNamedAsync(_service, _name, ..._args) {
        throw new Error('Not implemented');
    }
    resolveWithDependencies(service, dependencies) {
        // Import API container is passed as undefined because we only
        // care about carrying the dependencies
        const api = this._containerApiService.create(undefined);
        api.service(service)
            .dependencies(dependencies);
        return this.resolveWithDepBase(api);
    }
    resolveWithDependenciesAsync(_service, _dependencies) {
        throw new Error('Not implemented');
    }
    resolveWith(service) {
        const importApi = {
            execute: (api) => {
                let result;
                if (api.isDependenciesResolvable) {
                    result = this.resolveWithDepBase(api);
                }
                else {
                    const registration = this.createRegistration(api.serviceValue);
                    registration.name = api.nameValue;
                    registration.args = api.argsValue;
                    result = this.resolveBase(registration, api.throwResolveError);
                }
                if (result && api.cacheValue.use) {
                    this.addToCache(result, api);
                }
                return result;
            }
        };
        const api = this._containerApiService.create(importApi);
        return api.service(service);
    }
    registerImpl(registration) {
        registration.checkRegistrationType();
        registration.container = this;
        this._collection.addEntry(registration);
    }
    resolveBase(registration, throwIfNotFound) {
        const entry = this.resolveImpl(registration, throwIfNotFound);
        if (!entry && !throwIfNotFound) {
            return undefined;
        }
        if (this._pendingResolutions[entry.id]) {
            throw new CircularDependencyError(`${entry.service}`);
        }
        try {
            this._pendingResolutions[entry.id] = true;
            // ------------ with args always returns new instance ...
            if (registration.args && registration.args.length) {
                return this.createTrackable(entry, throwIfNotFound, registration.args);
            }
            return this.resolveScope(entry, throwIfNotFound);
        }
        finally {
            delete this._pendingResolutions[entry.id];
        }
    }
    resolveImpl(registration, throwIfNotFound) {
        const serviceEntry = this._collection.getEntry(registration);
        if (!serviceEntry && this._parent) {
            return this._parent.resolveImpl(registration, throwIfNotFound);
        }
        if (!serviceEntry && throwIfNotFound) {
            throw new ResolutionError({
                message: 'Could not resolve service',
                data: registration.service
            });
        }
        return serviceEntry;
    }
    resolveScope(registration, throwIfNotFound) {
        switch (registration.scope) {
            case scope.none:
                return this.createTrackable(registration, throwIfNotFound);
            case scope.container:
                return this.resolveContainerScope(registration, throwIfNotFound);
            case scope.hierarchy:
                return this.resolveHierarchyScope(registration, throwIfNotFound);
            default:
                throw new ResolutionError({ message: 'Unknown scoping' });
        }
    }
    resolveContainerScope(registration, throwIfNotFound) {
        let entry;
        if (registration.container !== this) {
            entry = registration.cloneFor(this);
            this._collection.addEntry(entry);
        }
        else {
            entry = registration;
        }
        if (!entry.instance) {
            entry.instance = this.createTrackable(entry, throwIfNotFound);
        }
        return entry.instance;
    }
    resolveHierarchyScope(registration, throwIfNotFound) {
        if (registration.container &&
            registration.container !== this) {
            const container = registration.container;
            return container.resolveBase(registration, throwIfNotFound);
        }
        if (!registration.instance) {
            registration.instance = this.createTrackable(registration, throwIfNotFound);
        }
        return registration.instance;
    }
    createTrackable(registration, throwIfNotFound, args) {
        try {
            let instance = this._invoker.invoke(registration, throwIfNotFound, args);
            if (registration.initializer) {
                instance = registration.initializer(this, instance);
            }
            if (registration.owner === owner.container &&
                !registration.isLazy &&
                registration.disposer) {
                this._disposableStorage.add(instance, registration.disposer);
            }
            return instance;
        }
        catch (error) {
            if (error instanceof CircularDependencyError) {
                throw error;
            }
            let message = 'Could not instantiate service';
            if (error instanceof ResolutionError) {
                message += `. ${error.message}`;
            }
            throw new ResolutionError({ message, data: registration.service, error });
        }
    }
    createRegistration(service) {
        return this._registrationBaseService.create(service);
    }
    createDependenciesRegistration(api) {
        const items = api.dependenciesValue.map(dependency => {
            if (!dependency.service) {
                throw new ResolutionError({ message: 'Service is not defined', data: dependency });
            }
            const registration = this.createRegistration(dependency.service);
            registration.copyDependency(dependency);
            registration.checkRegistrationType();
            const throwOnError = dependency.required !== false && api.throwResolveError;
            return {
                implementation: this.resolveImpl(registration, throwOnError),
                dependency
            };
        })
            .filter(item => item.implementation ||
            item.dependency.required === false ? true : false);
        if (items.length !== api.dependenciesValue.length)
            return [];
        return items.map(item => {
            const baseRegistration = item.dependency.required === false ?
                this.createRegistration(item.dependency.service)
                : item.implementation.cloneFor(this);
            baseRegistration.copyDependency(item.dependency);
            baseRegistration.disposer = undefined;
            baseRegistration.scope = this._dependencyScope;
            baseRegistration.owner = this._dependencyOwner;
            return baseRegistration;
        });
    }
    resolveWithDepBase(api) {
        const child = this.createChild();
        const registration = this.createRegistration(api.serviceValue);
        registration.args = api.argsValue;
        registration.name = api.nameValue;
        const implementation = this.resolveImpl(registration, api.throwResolveError);
        if (!implementation && !api.throwResolveError) {
            return undefined;
        }
        const baseRegistration = implementation.cloneFor(child);
        baseRegistration.args = api.argsValue;
        baseRegistration.name = api.nameValue;
        baseRegistration.disposer = undefined;
        baseRegistration.dependenciesValue = api.dependenciesValue;
        baseRegistration.scope = this._dependencyScope;
        baseRegistration.owner = this._dependencyOwner;
        const registrations = this.createDependenciesRegistration(api);
        if (registrations.length <= 0) {
            return undefined;
        }
        registrations.push(baseRegistration);
        child.add(registrations);
        return child.resolveBase(baseRegistration, api.throwResolveError);
    }
    createCache(resolutionCache) {
        return {
            get instance() {
                return resolutionCache.instance;
            },
            resolve(name) {
                return resolutionCache.resolve(name);
            }
        };
    }
    addToCache(value, api) {
        let name;
        if (api.cacheValue.name) {
            name = api.cacheValue.name;
        }
        else if (api.nameValue) {
            name = api.nameValue;
        }
        else if (api.serviceValue.name) {
            name = api.serviceValue.name;
        }
        else if (typeof api.serviceValue === 'string') {
            name = api.serviceValue;
        }
        else {
            throw new ResolutionError({ message: 'Missing cache name' });
        }
        this._resolutionCache.add(name, value);
    }
}

const decoratorResolutionParameter = Object.freeze({
    service: 1,
    value: 2,
    functionValue: 3
});

class Decorator {
    constructor(builderService, internalContainerService, _decoratorRegistrationApiService, _internalStorageService) {
        this._decoratorRegistrationApiService = _decoratorRegistrationApiService;
        this._internalStorageService = _internalStorageService;
        this._internalStorage = this._internalStorageService.create();
        internalContainerService.resolutionDetails = this._internalStorage;
        this._builder = builderService.create(internalContainerService);
    }
    build() {
        return this._builder.build();
    }
    provide(service) {
        const register = createRegister(this._builder);
        const api = this._decoratorRegistrationApiService.createRegistration(register);
        return api.provide(service);
    }
    provideSelf() {
        const register = createRegister(this._builder, true);
        const api = this._decoratorRegistrationApiService.createRegistration(register);
        return api.provideUndefined();
    }
    by(service) {
        const resolve = (api) => (target, _key, index) => {
            if (!api.service) {
                const dependencies = getMetadata(target);
                api.service = dependencies[index];
            }
            const bucket = this._internalStorage.register(target, () => ({}));
            bucket[index] = {
                service: api.service,
                args: api.args,
                attempt: api.attempt,
                name: api.name,
                cache: api.cache,
                type: decoratorResolutionParameter.service
            };
        };
        const api = this._decoratorRegistrationApiService.createResolution(resolve);
        return api.by(service);
    }
    resolveValue(value) {
        return (target, _key, index) => {
            const bucket = this._internalStorage.register(target, () => ({}));
            const type = isFunction(value) ?
                decoratorResolutionParameter.functionValue :
                decoratorResolutionParameter.value;
            bucket[index] = {
                value,
                type
            };
        };
    }
    register(service) {
        return this._builder.register(service);
    }
    import(builder) {
        this._builder.copy(builder);
    }
}
function createRegister(builder, isSelf = false) {
    return (api) => (target) => {
        if (!isPrototype(target)) {
            throw new DecoratorError({ message: 'Decorator target not supported, not a prototype', data: { target } });
        }
        const registration = isSelf ?
            builder.register(target).asSelf() :
            builder.register(api.service).asType(target);
        const initializer = api.initializedBy;
        if (initializer) {
            registration.initializeBy(initializer);
        }
        const isLazy = api.isLazy;
        if (isLazy) {
            registration.lazy();
        }
        const disposer = api.disposedBy;
        if (disposer) {
            registration.dispose(disposer);
        }
        const name = api.name;
        if (name) {
            registration.named(name);
        }
        const scope = api.scope || defaults.scope;
        registration.within(scope);
        return target;
    };
}

class RegistrationApi {
    constructor(_register) {
        this._register = _register;
        this.initializeBy = this.initializeBy.bind(this);
        this.lazy = this.lazy.bind(this);
        this.dispose = this.dispose.bind(this);
        this.named = this.named.bind(this);
        this.within = this.within.bind(this);
        this.transient = this.transient.bind(this);
        this.singleton = this.singleton.bind(this);
        this.instancePerContainer = this.instancePerContainer.bind(this);
        this.register = this.register.bind(this);
        this._isLazy = false;
    }
    get service() {
        return this._service;
    }
    get name() {
        return this._name;
    }
    get scope() {
        return this._scope;
    }
    get initializedBy() {
        return this._initializedBy;
    }
    get isLazy() {
        return this._isLazy;
    }
    get disposedBy() {
        return this._disposedBy;
    }
    provide(service) {
        checkNullArgument(service, 'service');
        const result = this.provideUndefined();
        this._service = service;
        return result;
    }
    provideUndefined() {
        this._service = undefined;
        return {
            initializeBy: this.initializeBy,
            lazy: this.lazy,
            dispose: this.dispose,
            named: this.named,
            within: this.within,
            transient: this.transient,
            singleton: this.singleton,
            instancePerContainer: this.instancePerContainer,
            register: this.register
        };
    }
    initializeBy(action) {
        checkNullArgument(action, 'action');
        this._initializedBy = action;
        return {
            lazy: this.lazy,
            dispose: this.dispose,
            named: this.named,
            within: this.within,
            transient: this.transient,
            singleton: this.singleton,
            instancePerContainer: this.instancePerContainer,
            register: this.register
        };
    }
    lazy() {
        this._isLazy = true;
        return {
            named: this.named,
            within: this.within,
            transient: this.transient,
            singleton: this.singleton,
            instancePerContainer: this.instancePerContainer,
            register: this.register
        };
    }
    dispose(action) {
        checkNullArgument(action, 'action');
        this._disposedBy = action;
        return {
            named: this.named,
            within: this.within,
            transient: this.transient,
            singleton: this.singleton,
            instancePerContainer: this.instancePerContainer,
            register: this.register
        };
    }
    named(name) {
        checkNullArgument(name, 'name');
        this._name = name;
        return {
            within: this.within,
            transient: this.transient,
            singleton: this.singleton,
            instancePerContainer: this.instancePerContainer,
            register: this.register
        };
    }
    within(scope) {
        checkNullArgument(scope, 'scope');
        this._scope = scope;
        return {
            register: this.register
        };
    }
    transient() {
        return this.within(scope.none);
    }
    singleton() {
        return this.within(scope.hierarchy);
    }
    instancePerContainer() {
        return this.within(scope.container);
    }
    register() {
        return this._register(this);
    }
}

class ResolutionApi {
    constructor(_resolve) {
        this._resolve = _resolve;
        this._attempt = false;
        this._args = [];
        this._cache = {
            use: false,
            name: undefined
        };
    }
    get service() {
        return this._service;
    }
    set service(value) {
        this._service = value;
    }
    get args() {
        return this._args;
    }
    get name() {
        return this._name;
    }
    get attempt() {
        return this._attempt;
    }
    get cache() {
        return this._cache;
    }
    by(service) {
        this._service = service;
        return {
            args: this.argsAction.bind(this),
            attempt: this.attemptAction.bind(this),
            name: this.nameAction.bind(this),
            cache: this.cacheAction.bind(this),
            resolve: this.resolveAction.bind(this)
        };
    }
    argsAction(...value) {
        this._args = value;
        return {
            attempt: this.attemptAction.bind(this),
            name: this.nameAction.bind(this),
            cache: this.cacheAction.bind(this),
            resolve: this.resolveAction.bind(this)
        };
    }
    attemptAction() {
        this._attempt = true;
        return {
            name: this.nameAction.bind(this),
            cache: this.cacheAction.bind(this),
            resolve: this.resolveAction.bind(this)
        };
    }
    nameAction(value) {
        this._name = value;
        return {
            cache: this.cacheAction.bind(this),
            resolve: this.resolveAction.bind(this)
        };
    }
    cacheAction(name) {
        this._cache = {
            use: true,
            name
        };
        return {
            resolve: this.resolveAction.bind(this)
        };
    }
    resolveAction() {
        return this._resolve(this);
    }
}

class Invoker {
    constructor(_container, _resolutionDetails) {
        this._container = _container;
        this._resolutionDetails = _resolutionDetails;
    }
    invoke(registration, throwIfNotFound, args) {
        if (registration.isLazy) {
            return () => this.invokeInternal(registration, throwIfNotFound, args);
        }
        return this.invokeInternal(registration, throwIfNotFound, args);
    }
    invokeInternal(registration, throwIfNotFound, args) {
        switch (registration.registrationType) {
            case registrationType.factoryType:
                return this.instantiate(registration.factoryType, registration, throwIfNotFound, args);
            case registrationType.factoryValue:
                return registration.factoryValue;
            case registrationType.factory:
            default:
                return this.createByFactory(registration, args);
        }
    }
    createByFactory(registration, args = []) {
        return registration.factory(registration.container, ...args);
    }
    instantiate(type, registration, throwIfNotFound, args) {
        if (args && args.length &&
            registration.params.length) {
            throw new ResolutionError({
                message: [
                    'Could not instantiate type.',
                    'Arguments and dependencies are not allowed for simultaneous resolution.',
                    'Pick dependencies or arguments'
                ].join(' '),
                data: type
            });
        }
        if (args && args.length) {
            return construct(type, args);
        }
        if (registration.params.length) {
            return this.instantiateByParams(type, registration, throwIfNotFound);
        }
        const dependencies = this.getDependencies(type);
        return this.instantiateByDependencies(type, dependencies);
    }
    instantiateByParams(type, registration, throwIfNotFound) {
        const params = registration.params
            .map(item => {
            const dependency = registration.dependenciesValue.filter(d => d.service === item)[0];
            const dependencyName = dependency ? dependency.named : null;
            if (throwIfNotFound) {
                return !!dependencyName ?
                    this._container.resolveNamed(item, dependencyName) :
                    this._container.resolve(item);
            }
            return !!dependencyName ?
                this._container.tryResolveNamed(item, dependencyName) :
                this._container.tryResolve(item);
        })
            .filter(item => item);
        return construct(type, params);
    }
    instantiateByDependencies(type, dependencies) {
        const params = dependencies
            .map((dependency, index) => {
            const depParams = this._resolutionDetails ? this._resolutionDetails.tryGet(type) : null;
            const depParamsValue = depParams ? depParams[index] : null;
            if (!depParamsValue) {
                return this._container.resolve(dependency);
            }
            if (depParamsValue.type === decoratorResolutionParameter.value) {
                return depParamsValue.value;
            }
            if (depParamsValue.type === decoratorResolutionParameter.functionValue) {
                return depParamsValue.value();
            }
            const resolutionItem = depParamsValue.service || dependency;
            const resolution = this._container.resolveWith(resolutionItem);
            if (depParamsValue.args && depParamsValue.args.length) {
                resolution.args(...depParamsValue.args);
            }
            if (depParamsValue.name) {
                resolution.name(depParamsValue.name);
            }
            if (depParamsValue.attempt === true) {
                resolution.attempt();
            }
            if (depParamsValue.cache && depParamsValue.cache.use) {
                resolution.cache(depParamsValue.cache.name);
            }
            return resolution.exec();
        });
        return construct(type, params);
    }
    getDependencies(type) {
        return getMetadata(type);
    }
}

class SubstituteStorage {
    constructor() {
        this._known = Object.create(null);
        this._unknown = { head: null, tail: null };
    }
    add(value) {
        if (value.method) {
            this.addKnownSubstitute(value);
        }
        else {
            this.addUnknownSubstitute(value);
        }
    }
    getSubstitutes(name, types) {
        const unknown = this._unknown.head ?
            this.copyList(this._unknown, (value) => !value.type || types.indexOf(value.type) >= 0)
            :
                null;
        const bucket = this._known[name];
        if (!bucket) {
            return unknown ? unknown.head : null;
        }
        const result = this.copyList(bucket, (value) => {
            if (!value.type || types.indexOf(value.type) >= 0) {
                return true;
            }
            throw this.createIncompatibleTypeError(value.type, types);
        });
        result.tail.next = (unknown ? unknown.head : result.tail.next) || undefined;
        return result.head;
    }
    addKnownSubstitute(value) {
        const method = value.method;
        const bucket = this._known[method] || { head: null, tail: null };
        this.addToList(value, bucket);
        this._known[method] = bucket;
    }
    addUnknownSubstitute(value) {
        this.addToList(value, this._unknown);
    }
    createIncompatibleTypeError(type, types) {
        const typeMessage = types.map(item => this.callTypeToString(item)).join(', ');
        const message = [
            'Could not match proxy type and property type.',
            `Expected one of: ${typeMessage}.`,
            `Actual: ${this.callTypeToString(type)}`
        ].join(' ');
        throw new ProxyError({ message });
    }
    addToList(value, list) {
        if (!list.head) {
            list.head = value;
            list.tail = value;
        }
        else {
            list.tail.next = value;
            list.tail = value;
        }
    }
    copyList(list, predicate) {
        const result = {
            head: null,
            tail: null
        };
        let current = list.head;
        while (current) {
            if (predicate(current)) {
                const substitute = {
                    method: current.method,
                    type: current.type,
                    wrapper: current.wrapper
                };
                this.addToList(substitute, result);
            }
            current = current.next || null;
        }
        return result;
    }
    callTypeToString(type) {
        switch (type) {
            case callInfo.field: return 'Field';
            case callInfo.getter: return 'Getter';
            case callInfo.setter: return 'Setter';
            case callInfo.getterSetter: return 'GetterSetter';
            case callInfo.method: return 'Method';
            default:
                return 'Any';
        }
    }
}

class Interceptor {
    constructor(_proxy) {
        this._proxy = _proxy;
    }
    interceptPrototype(subject, substitutes) {
        this.assertPrototype(subject);
        const storage = this.convertParams(subject, substitutes);
        return this._proxy.byPrototype(subject, storage);
    }
    interceptInstance(subject, substitutes) {
        this.assertObject(subject);
        const storage = this.convertParams(subject, substitutes);
        return this._proxy.byInstance(subject, storage);
    }
    intercept(subject, substitutes) {
        const storage = this.convertParams(subject, substitutes);
        return this.convertToIntercept(subject, storage);
    }
    withSubstitute(substitute) {
        const storage = new SubstituteStorage();
        const interceptInstance = (subject) => {
            this.assertObject(subject);
            return this._proxy.byInstance(subject, storage);
        };
        const interceptPrototype = (subject) => {
            this.assertPrototype(subject);
            return this._proxy.byPrototype(subject, storage);
        };
        const intercept = (subject) => {
            return this.convertToIntercept(subject, storage);
        };
        return this.with(interceptInstance, interceptPrototype, intercept, storage, substitute);
    }
    with(interceptInstance, interceptPrototype, intercept, storage, substitute) {
        storage.add(this.createSubstitute(substitute));
        return {
            withSubstitute: this.with.bind(this, interceptInstance, interceptPrototype, intercept, storage),
            interceptInstance,
            interceptPrototype,
            intercept
        };
    }
    convertParams(subject, substitutes) {
        checkNullArgument(subject, 'subject');
        if (!substitutes) {
            return undefined;
        }
        const data = isArray(substitutes) ?
            substitutes :
            [substitutes];
        return this.transformSubstitutes(data);
    }
    convertToIntercept(subject, storage) {
        let result;
        if (isPrototype(subject)) {
            result = this._proxy.byPrototype(subject, storage);
        }
        else if (isObject(subject)) {
            result = this._proxy.byInstance(subject, storage);
        }
        else {
            throw new ArgumentError('subject', {
                message: 'Subject should be a prototype function or an object'
            });
        }
        return result;
    }
    transformSubstitutes(substitutes) {
        const storage = new SubstituteStorage();
        return substitutes.reduce((storage, current) => {
            const substitute = this.createSubstitute(current);
            storage.add(substitute);
            return storage;
        }, storage);
    }
    createSubstitute(value) {
        if (!value.wrapper) {
            throw new ArgumentError('wrapper', {
                message: 'Missing interceptor wrapper', data: value
            });
        }
        return {
            method: value.method,
            type: value.type || callInfo.any,
            wrapper: value.wrapper,
            next: undefined
        };
    }
    assertPrototype(subject) {
        if (!isPrototype(subject)) {
            throw new ArgumentError('subject', {
                message: 'Subject should be a prototype function'
            });
        }
    }
    assertObject(subject) {
        if (isPrototype(subject)) {
            throw new ArgumentError('subject', { message: 'Subject should be an object' });
        }
    }
}

const propertyType = Object.freeze({
    method: 1,
    getter: 2,
    setter: 3,
    fullProperty: 4,
    field: 5 // field
});

class Decorator$1 {
    wrap(strategyInfo) {
        strategyInfo = this.copyStrategy(strategyInfo);
        const strategyStore = strategyInfo.substitute ?
            this.defineWrapStrategies() :
            this.defineNonWrapStrategies();
        const strategy = strategyStore[strategyInfo.type];
        strategy(strategyInfo);
    }
    defineNonWrapStrategies() {
        const result = {};
        result[propertyType.method] = (strategyInfo) => {
            const value = strategyInfo.source[strategyInfo.name];
            strategyInfo.destination[strategyInfo.name] = function () {
                const args = Array.prototype.slice.call(arguments, 0);
                return value.apply(this, args);
            };
        };
        result[propertyType.getter] = (strategyInfo) => {
            const { configurable, enumerable } = strategyInfo.descriptor ?
                strategyInfo.descriptor : { configurable: true, enumerable: true };
            Object.defineProperty(strategyInfo.destination, strategyInfo.name, {
                get: this.defineGetter(strategyInfo),
                configurable,
                enumerable
            });
        };
        result[propertyType.setter] = (strategyInfo) => {
            const { configurable, enumerable } = strategyInfo.descriptor ?
                strategyInfo.descriptor : { configurable: true, enumerable: true };
            Object.defineProperty(strategyInfo.destination, strategyInfo.name, {
                set: this.defineSetter(strategyInfo),
                configurable,
                enumerable
            });
        };
        result[propertyType.fullProperty] = (strategyInfo) => {
            const { configurable, enumerable } = strategyInfo.descriptor ?
                strategyInfo.descriptor : { configurable: true, enumerable: true };
            Object.defineProperty(strategyInfo.destination, strategyInfo.name, {
                get: this.defineGetter(strategyInfo),
                set: this.defineSetter(strategyInfo),
                configurable,
                enumerable
            });
        };
        result[propertyType.field] = result[propertyType.fullProperty];
        return result;
    }
    defineWrapStrategies() {
        const result = {};
        const createCallChainFromList = this.createCallChainFromList.bind(this);
        const defineWrapGetter = this.defineWrapGetter.bind(this);
        const defineWrapSetter = this.defineWrapSetter.bind(this);
        const defineGetter = this.defineGetter.bind(this);
        const defineSetter = this.defineSetter.bind(this);
        result[propertyType.method] = (strategyInfo) => {
            const value = strategyInfo.source[strategyInfo.name];
            strategyInfo.destination[strategyInfo.name] = function () {
                const args = Array.prototype.slice.call(arguments, 0);
                const delegate = (args) => {
                    if (!args || !isArray(args)) {
                        args = [args];
                    }
                    return value.apply(this, args);
                };
                return createCallChainFromList({
                    delegate,
                    strategyInfo,
                    args: ImmutableArray.createImmutable(args),
                    wrapperContext: this,
                });
            };
        };
        result[propertyType.getter] = (strategyInfo) => {
            Object.defineProperty(strategyInfo.destination, strategyInfo.name, {
                get: defineWrapGetter(strategyInfo),
                configurable: true,
                enumerable: strategyInfo.descriptor ? strategyInfo.descriptor.enumerable : true
            });
        };
        result[propertyType.setter] = (strategyInfo) => {
            Object.defineProperty(strategyInfo.destination, strategyInfo.name, {
                set: defineWrapSetter(strategyInfo),
                configurable: true,
                enumerable: strategyInfo.descriptor ? strategyInfo.descriptor.enumerable : true
            });
        };
        result[propertyType.fullProperty] = (strategyInfo) => {
            const getter = strategyInfo.substitute.type === callInfo.any ||
                strategyInfo.substitute.type === callInfo.getterSetter ||
                strategyInfo.substitute.type === callInfo.getter ||
                strategyInfo.substitute.type === callInfo.field ?
                defineWrapGetter(strategyInfo) : defineGetter(strategyInfo);
            const setter = strategyInfo.substitute.type === callInfo.any ||
                strategyInfo.substitute.type === callInfo.getterSetter ||
                strategyInfo.substitute.type === callInfo.setter ||
                strategyInfo.substitute.type === callInfo.field ?
                defineWrapSetter(strategyInfo) : defineSetter(strategyInfo);
            Object.defineProperty(strategyInfo.destination, strategyInfo.name, {
                get: getter,
                set: setter,
                configurable: true,
                enumerable: strategyInfo.descriptor ? strategyInfo.descriptor.enumerable : true
            });
        };
        result[propertyType.field] = result[propertyType.fullProperty];
        return result;
    }
    defineWrapSetter(strategyInfo) {
        const createCallChainFromList = this.createCallChainFromList.bind(this);
        const defineSetter = this.defineSetter.bind(this);
        return function (value) {
            const delegate = defineSetter(strategyInfo).bind(this);
            return createCallChainFromList({
                delegate,
                strategyInfo,
                args: ImmutableArray.createImmutable([value]),
                wrapperContext: this,
                callType: callInfo.setter,
            });
        };
    }
    defineWrapGetter(strategyInfo) {
        const createCallChainFromList = this.createCallChainFromList.bind(this);
        const defineGetter = this.defineGetter.bind(this);
        return function () {
            const delegate = defineGetter(strategyInfo).bind(this);
            return createCallChainFromList({
                delegate,
                strategyInfo,
                args: ImmutableArray.createImmutable([]),
                wrapperContext: this,
                callType: callInfo.getter,
            });
        };
    }
    defineGetter(strategyInfo) {
        return function () {
            return strategyInfo.contextName ? this[strategyInfo.contextName][strategyInfo.name]
                : strategyInfo.source[strategyInfo.name];
        };
    }
    defineSetter(strategyInfo) {
        return function (argValue) {
            if (strategyInfo.contextName) {
                this[strategyInfo.contextName][strategyInfo.name] = argValue;
            }
            else {
                strategyInfo.source[strategyInfo.name] = argValue;
            }
        };
    }
    createCallChainFromList(info) {
        const mainCallInfo = this.createCallInfo(info);
        this.createCallAction(mainCallInfo, info, info.strategyInfo.substitute.next);
        return info.strategyInfo.substitute.wrapper.call(info.wrapperContext, mainCallInfo);
    }
    createCallAction(callInfo, info, substitute) {
        if (!substitute) {
            return;
        }
        const createCallAction = this.createCallAction.bind(this);
        const childCallInfo = this.createCallInfo(info);
        const wrapper = substitute.wrapper.bind(info.wrapperContext);
        callInfo.next = result => {
            childCallInfo.result = result;
            createCallAction(childCallInfo, info, substitute.next);
            return wrapper(childCallInfo);
        };
    }
    createCallInfo(info) {
        const getter = info.delegate;
        const setter = info.delegate;
        return {
            source: info.strategyInfo.source,
            name: info.strategyInfo.name,
            args: info.args.value,
            type: info.callType || info.strategyInfo.substitute.type,
            invoke: info.delegate,
            get: info.callType === callInfo.getter ? getter : undefined,
            set: info.callType === callInfo.setter ? setter : undefined
        };
    }
    copyStrategy(strategyInfo) {
        return {
            type: strategyInfo.type,
            descriptor: strategyInfo.descriptor,
            substitute: strategyInfo.substitute,
            name: strategyInfo.name,
            source: strategyInfo.source,
            destination: strategyInfo.destination,
            contextName: strategyInfo.contextName
        };
    }
}

const blackListProperties = [
    '__lookupGetter__',
    '__lookupSetter__',
    '__proto__',
    '__defineGetter__',
    '__defineSetter__',
    'hasOwnProperty',
    'propertyIsEnumerable',
    'constructor'
];
const isBlackListProperty = (property) => {
    return blackListProperties.indexOf(property) >= 0;
};
const getPropertyType = (descriptor) => {
    if (!descriptor) {
        return propertyType.field;
    }
    if (descriptor.value && isFunction(descriptor.value)) {
        return propertyType.method;
    }
    if (descriptor.get && !descriptor.set)
        return propertyType.getter;
    if (!descriptor.get && descriptor.set)
        return propertyType.setter;
    if (descriptor.get && descriptor.set)
        return propertyType.fullProperty;
    return propertyType.field;
};
const createProxy = (self, parent, storage) => {
    // tslint:disable-next-line: function-name
    function Proxy() {
        this._parent = construct(parent, arguments);
        Object.getOwnPropertyNames(this._parent)
            .filter(name => !isBlackListProperty(name))
            .filter(name => !(name in this) && !(name in Proxy.prototype))
            .map(p => self.createStrategyInfo(this._parent, this, p))
            .forEach(s => self.decorateProperty(s, storage));
    }
    return Proxy;
};

class Proxy {
    constructor(_decorator) {
        this._decorator = _decorator;
        this.restrictedProperties = getAllPropertyNames(Function);
        this.propTypeToDescriptor = {
            [propertyType.method]: [callInfo.any, callInfo.method],
            [propertyType.getter]: [callInfo.any, callInfo.getter],
            [propertyType.setter]: [callInfo.any, callInfo.setter],
            [propertyType.fullProperty]: [
                callInfo.any,
                callInfo.getter,
                callInfo.setter,
                callInfo.getterSetter
            ],
            [propertyType.field]: [
                callInfo.any,
                callInfo.getter,
                callInfo.setter,
                callInfo.getterSetter,
                callInfo.field
            ]
        };
    }
    byPrototype(parent, storage) {
        const source = parent.prototype;
        const proxy = createProxy(this, parent, storage);
        getAllPropertyNames(source)
            .filter(name => !isBlackListProperty(name))
            .map(p => this.createStrategyInfo(source, proxy.prototype, p, '_parent'))
            .forEach(s => this.decorateProperty(s, storage));
        Object.getOwnPropertyNames(parent)
            .filter(name => this.restrictedProperties.indexOf(name) === -1)
            .map(p => this.createStrategyInfo(parent, proxy, p))
            .forEach(s => this.decorateProperty(s, storage));
        return proxy;
    }
    byInstance(parent, storage) {
        const result = Object.create({});
        getAllPropertyNames(parent)
            .filter(name => !isBlackListProperty(name))
            .map(p => this.createStrategyInfo(parent, result, p))
            .forEach(s => this.decorateProperty(s, storage));
        return result;
    }
    decorateProperty(strategyInfo, storage) {
        if (storage) {
            const types = this.propTypeToDescriptor[strategyInfo.type];
            const substitute = storage.getSubstitutes(strategyInfo.name, types);
            strategyInfo.substitute = substitute || undefined;
            this._decorator.wrap(strategyInfo);
            return;
        }
        this._decorator.wrap(strategyInfo);
    }
    createStrategyInfo(source, destination, name, contextName) {
        const descriptor = getPropertyDescriptor(source, name);
        const propertyType = getPropertyType(descriptor);
        return {
            type: propertyType,
            descriptor,
            substitute: undefined,
            name,
            source: source,
            destination: destination,
            contextName
        };
    }
}

class Scaffold {
    createBuilder() {
        const internalContainerService = this.internalContainerService();
        return this.builderService().create(internalContainerService);
    }
    createDecorator() {
        const decoratorRegistrationService = this.decoratorRegistrationApiService();
        const internalStorageService = this.internalStorageService();
        const internalContainerService = this.internalContainerService();
        return new Decorator(this.builderService(), internalContainerService, decoratorRegistrationService, internalStorageService);
    }
    createInterceptor() {
        const decorator = new Decorator$1();
        const proxy = new Proxy(decorator);
        return new Interceptor(proxy);
    }
    builderService() {
        return {
            create: (internalContainerService) => {
                const baseRegistrationService = this.registrationBaseService();
                const instanceRegistrationService = this.instanceRegistrationService();
                const containerService = this.containerService();
                return new ContainerBuilder(baseRegistrationService, instanceRegistrationService, internalContainerService, containerService);
            }
        };
    }
    internalStorageService() {
        return {
            create() {
                return new InternalStorage();
            }
        };
    }
    registrationStorageService() {
        const service = {
            create() {
                return new InternalStorage();
            }
        };
        return {
            create() {
                return new RegistrationStorage(service);
            }
        };
    }
    disposableStorageService() {
        return {
            create() {
                return new DisposableStorage();
            }
        };
    }
    resolutionCacheService() {
        return {
            create() {
                return new ResolutionCache();
            }
        };
    }
    registrationBaseService() {
        return {
            create(service) {
                return new RegistrationBase(service);
            }
        };
    }
    containerService() {
        return {
            create: container
        };
    }
    containerApiService() {
        return {
            create: (container) => {
                return new Api(container);
            }
        };
    }
    instanceRegistrationService() {
        return {
            create: (baseRegistration) => {
                return new Registration(baseRegistration);
            }
        };
    }
    internalContainerService() {
        const dependencies = () => {
            const registrationStorageService = this.registrationStorageService();
            const disposableStorageService = this.disposableStorageService();
            const baseRegistrationService = this.registrationBaseService();
            const containerApiService = this.containerApiService();
            const invokerService = this.invokerService();
            const resolutionCacheService = this.resolutionCacheService();
            return {
                registrationStorageService,
                disposableStorageService,
                baseRegistrationService,
                containerApiService,
                invokerService,
                resolutionCacheService
            };
        };
        return {
            create() {
                const params = dependencies();
                return new InternalContainer(params.registrationStorageService, params.disposableStorageService, params.baseRegistrationService, params.containerApiService, params.invokerService, params.resolutionCacheService, this.resolutionDetails);
            }
        };
    }
    invokerService() {
        return {
            create(container, resolutionDetails) {
                return new Invoker(container, resolutionDetails);
            }
        };
    }
    decoratorRegistrationApiService() {
        return {
            createRegistration(register) {
                return new RegistrationApi(register);
            },
            createResolution(resolve) {
                return new ResolutionApi(resolve);
            }
        };
    }
}

/**
 * Dependency injection container for TypeScript / JavaScript
 *
 * @packageDocumentation
 */
const scaffold = new Scaffold();
/**
 * Creates an instance on {@link IContainerBuilder} interface
 * @public
 * @returns - an instance on {@link IContainerBuilder} interface
 */
const builder = () => scaffold.createBuilder();
/**
 * Creates an instance on {@link IDecorator} interface
 * @public
 * @returns - an instance on {@link IDecorator} interface
 */
const decorator = () => scaffold.createDecorator();
/**
 * Creates an instance on {@link IInterceptor} interface
 * @public
 * @returns - an instance on {@link IInterceptor} interface
 */
const interceptor = () => scaffold.createInterceptor();

export { ApplicationError, ArgumentError, CircularDependencyError, DecoratorError, ProxyError, ResolutionError, builder, callInfo, decorator, interceptor, scope };
//# sourceMappingURL=index.js.map
